from swimlane_platform.lib.version_manager import semver_parse
from swimlane_platform.lib import BaseWithLog, \
    Configuration, \
    ValidationException, \
    names, \
    DockerManager, \
    AnswerRequiredValidator, \
    DockerComposeManager
from swimlane_platform.lib.constants import SWIMLANE_UID_GID
from swimlane_platform.lib.debug_decorators import debug_function_return, info_function_start_finish
from os import getuid, path, mkdir, rmdir
from json import load
from swimlane_platform.lib.os_manager import OsManager
from uuid import uuid1


class PreFlightCheck(BaseWithLog):
    """
    Class responsible for checking environment before attempting to install.
    """

    def __init__(self, config):
        # type: (Configuration) -> None
        super(PreFlightCheck, self).__init__(config)
        self.os_manager = OsManager(self.logger)

    @info_function_start_finish('Pre-flight checks.')
    def run(self):
        # type: () -> bool
        verification_result = True
        try:
            verification_result &= self.check_os()
            verification_result &= self.check_current_user()
            verification_result &= self.check_docker_version()
            verification_result &= self.check_docker_compose_version()
            verification_result &= self.check_docker_repository_access()
            verification_result &= self.check_installation_dir_permissions()
        except ValidationException as e:
            self.logger.error('Critical validation failed and further checks will not continue: {error}'
                              .format(error=e))
            verification_result = False
        return verification_result

    @debug_function_return
    def check_os(self):
        # type: () -> bool
        """
        Validates if Swimlane is getting installed on supported os.
        """
        if not self.os_manager.is_linux():
            raise ValidationException('Operating system must be linux.')
        match_centos = self.os_manager.is_match_distribution_linux('centos', '7.5.1804')
        match_rhel = self.os_manager.is_match_distribution_linux('redhat', '7.6.0', True)
        if not (match_centos or match_rhel):
            _os = self.os_manager.get_platform()
            raise ValidationException('{os} is not on the list of supported distributions.'.format(os=_os))
        return True

    @debug_function_return
    def check_current_user(self):
        # type: () -> bool
        """
        Checks if user is not a service user, but sudo.
        """
        uid = getuid()
        if uid == 0:
            self.logger.warn('Installer is currently running as root (sudo). This is NOT RECOMMENDED!')
            questions = [{
                'type': 'confirm',
                'name': 'continue_if_user_is_root',
                'message': 'Do you still wish to continue with root user?'
            }]
            self.config.collect(questions)
            if not self.config.args.continue_if_user_is_root:
                raise ValidationException("User is root, installation aborted.")
        elif uid != SWIMLANE_UID_GID:
            self.logger.warn('Installer is currently running as a user with a UID of {uid}. '.format(uid=uid)
                            + 'It is recommended that you run Swimlane under a user with a UID of '
                            + '{uid_gid} and a GID of {uid_gid}, otherwise Swimlane might not start '.format(uid_gid=SWIMLANE_UID_GID)
                            + 'once installation is complete.')
            questions = [{
                'type': 'confirm',
                'name': 'continue_if_user_has_incorrect_uid_gid',
                'message': 'Do you still wish to continue the install as the current user?'
            }]
            self.config.collect(questions)
            if not self.config.args.continue_if_user_has_incorrect_uid_gid:
                raise ValidationException('User does not have correct UID or GID, aborting')
        return True

    # noinspection PyBroadException
    @debug_function_return
    def check_docker_version(self):
        # type: () -> bool
        """
        Verifies the installed docker meets requirements.
        """
        # skip verification on RedHat - it uses it's own docker.
        if self.os_manager.get_distribution_linux() == 'rhel':
            return True

        version = '18.6.0'
        try:
            docker_manager = DockerManager(self.logger)
        except Exception:
            self.logger.error('Docker not found. '
                              'Please install Docker version ${version} or greater.'.format(version=version))
            return False
        if docker_manager.docker_version() < semver_parse(version):
            self.logger.error('Unsupported version of Docker detected. '
                              'Please install Docker version ${version} or greater.'.format(version=version))
            return False
        try:
            _ = docker_manager.images
        except Exception:
            self.logger.error('Installer cannot run Docker commands. '
                              'Please ensure that docker is running and a user is in the docker group.')
            return False
        return True

    # noinspection PyBroadException
    @debug_function_return
    def check_docker_compose_version(self):
        # type: () -> bool
        """
        Verifies the installed docker-compose meets requirements.
        """
        version = '1.22.0'
        try:
            docker_compose_version = DockerComposeManager(self.logger, '').docker_compose_version()
            if docker_compose_version < semver_parse(version):
                self.logger.error('docker-compose version is unsupported. '
                                  'Please install docker-compose version ${version} or greater.'
                                  .format(version=version))
                return False
            return True
        except Exception:
            self.logger.error('docker-compose not found. '
                              'Please install docker-compose version ${version} or greater.'
                              .format(version=version))
            return False

    @debug_function_return
    def check_docker_repository_access(self):
        # type: () -> bool
        """
        Checks if docker can login to nexus or PyPi.
        :return: True if passes
        """
        if self.config.args.installer_is_offline:
            return True
        repo_name = names.DEV_REPOSITORY.rstrip('/') if self.config.args.dev else 'https://index.docker.io/v1/'
        config_json = path.expanduser('~/.docker/config.json')
        if not path.exists(config_json):
            self.logger.debug("{name} is not found so no docker information.".format(name=config_json))
            no_login = True
        else:
            with open(path.expanduser(config_json)) as fs:
                docker_config = load(fs)
            registered_repositories = [str(key) for key, value in docker_config['auths'].items()]
            no_login = not [1 for repository in registered_repositories if repo_name in repository]
        if no_login:
            questions = [
                {
                    'type': 'confirm',
                    'name': 'collect_docker_login',
                    'message': 'You are not logged in to {name}, log in now?'.format(name=repo_name)
                },
                {
                    'type': 'input',
                    'name': 'docker_repo_user',
                    'message': '{name} user name?'.format(name=repo_name),
                    'when': lambda a: 'collect_docker_login' in a and a['collect_docker_login'],
                    'validate': AnswerRequiredValidator
                },
                {
                    'type': 'password',
                    'name': 'docker_repo_password',
                    'message': '{name} password?'.format(name=repo_name),
                    'when': lambda a: 'collect_docker_login' in a and a['collect_docker_login'],
                    'validate': AnswerRequiredValidator
                }]
            self.config.collect(questions)
            if not self.config.args.docker_repo_user or not self.config.args.docker_repo_password:
                self.logger.error("Cannot proceed without login to {name}.".format(name=repo_name))
                return False
            DockerManager(self.logger).login(repo_name, self.config.args.docker_repo_user,
                                             self.config.args.docker_repo_password)
        return True

    # noinspection PyBroadException
    @debug_function_return
    def check_installation_dir_permissions(self):
        # type: () -> bool
        """
        Checks if user can write to /opt
        :return: True if can.
        """
        dir_id = 'f' + str(uuid1()).replace('-', '_')
        new_dir = path.join(names.INSTALL_DIR, dir_id)
        try:
            mkdir(new_dir)
            rmdir(new_dir)
        except Exception:
            self.logger.error('Cannot install to {name}. Ensure current user has write permissions.'
                              .format(name=names.INSTALL_DIR))
            return False
        return True



from swimlane_platform.lib.mongo_manager import MongoManager
from swimlane_platform.lib.names import names
from swimlane_platform.lib.debug_decorators import info_function_start_finish, debug_function_args
from swimlane_platform.lib.docker_compose_manager import DockerComposeManager
from swimlane_platform.shared_steps import enable_turbine, set_mongo_tls_flags
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
import semver
from os import path


class UpgradeFrom912To1000(UpgradeStep):
    FROM = semver.parse_version_info('9.1.2')  # type: semver.VersionInfo
    TO = semver.parse_version_info('10.0.0')  # type: semver.VersionInfo

    @debug_function_args
    def upgrade_image_versions(self, install_dir, dev):
        # type: (str, bool) -> None
        """
        Changes image versions to the new ones
        :param dev: If the images will be pulled from development repository.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        docker_compose = self.upgrade_standard_images(dev, install_dir)
        docker_compose.set('mongo:4.2.1', 'services', names.SW_MONGO, 'image')
        docker_compose.save()

    @debug_function_args
    def upgrade_compatibility_mode(self):
        # type: () -> None
        """
        Issues database compatibility command.
        """
        docker_compose_file = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE)
        docker_compose_manager = DockerComposeManager(self.logger, docker_compose_file)
        docker_compose_manager.docker_compose_up(names.SW_MONGO)
        MongoManager(self.config).set_database_compatibility('4.2')
        docker_compose_manager.docker_compose_down()

    @info_function_start_finish('Upgrade From 9.1.2 To 10.0.0.')
    def process(self):
        # type: () -> None
        enable_turbine.run(self.config)
        self.upgrade_image_versions(names.INSTALL_DIR, self.config.args.dev)
        set_mongo_tls_flags.run(self.config)
        self.upgrade_compatibility_mode()



from swimlane_platform.lib import names
from swimlane_platform.lib.debug_decorators import info_function_start_finish, debug_function_args
from swimlane_platform.shared_steps import enable_turbine
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
import semver


class UpgradeFrom1040rc2To1040rc3(UpgradeStep):
    FROM = semver.parse_version_info('10.4.0-rc2')  # type: semver.VersionInfo
    TO = semver.parse_version_info('10.4.0-rc3')  # type: semver.VersionInfo

    @info_function_start_finish('Upgrade From 10.4.0-rc2 To 10.4.0-rc3')
    def process(self):
        # type: () -> None
        enable_turbine.run(self.config)
        self.upgrade_image_versions(names.INSTALL_DIR, self.config.args.dev)

    @debug_function_args
    def upgrade_image_versions(self, install_dir, dev):
        # type: (str, bool) -> None
        """
        Updates image versions
        :param dev: If the images will be pulled from development repository.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        docker_compose = self.upgrade_standard_images(self.config.args.dev, names.INSTALL_DIR)
        docker_compose.save()



from swimlane_platform.lib.logger import SplitStreamLogger
from swimlane_platform.lib.args_config_questions import Configuration
from swimlane_platform.lib.names import names


class BaseWithLog(object):

    def __init__(self, config):
        # type: (Configuration) -> None
        self.config = config
        verbose = config.args.verbose if config.args.verbose else 0
        name = self.__class__.__name__ if self.__class__ else names.APP_NAME
        self.logger = SplitStreamLogger(name, verbose, config.args.log)



from typing import Union
from os import path
from json import load
from swimlane_platform.lib.mongo_manager import MongoManager
from swimlane_platform.lib.functions import find_remote_sibling, copy_if_not_exist
from swimlane_platform.lib import names, DockerImage, DockerComposeManager, turbine_enable_questions
from swimlane_platform.lib.env_manager import read_dict, write_dict
from swimlane_platform.lib.args_config_questions import Configuration
from swimlane_platform.lib.base_with_log import BaseWithLog
from swimlane_platform.lib.docker_compose_file_manager import DockerComposeFileManager
from swimlane_platform.lib.debug_decorators import info_function_start_finish, debug_function_return
from swimlane_platform.lib.db_configuration import DbConfiguration


class EnableTurbine(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> None
        super(EnableTurbine, self).__init__(config)
        self.secrets_dir = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)
        self.template_dir = find_remote_sibling(path.dirname(path.realpath(__file__)), names.TEMPLATE_DIR)
        self.mongo_manager = MongoManager(self.config)
        self.db_configuration = DbConfiguration(self.config)

    @info_function_start_finish('Enabling Turbine.')
    def run(self):
        # type: () -> None
        """
        Main method for enabling turbine engine.
        """
        self.copy_files()
        self.modify_env_file()
        self.modify_compose_file()
        self.initialize_turbine_database()

    @info_function_start_finish('Adding turbine compose file to main docker-compose.')
    def modify_compose_file(self, main=None, turbine=None):
        # type: (Union[None, str], Union[None, str]) -> None
        """
        Overrides (in a docker-compose sense) main docker-compose file with Turbine.
        :param main: Main docker-compose file path (for testing)
        :param turbine: Turbine docker-compose file path (for testing)
        """
        main = main if main else path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE)
        turbine = turbine if turbine else path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_TURBINE_FILE)
        docker_compose_main = DockerComposeFileManager(self.logger, main)
        docker_compose_turbine = DockerComposeFileManager(self.logger, turbine)
        version = self.config.args.turbine_image_version if self.config.args.dev \
            else self.get_turbine_version_from_json()
        for name, service in docker_compose_turbine.get('services').items():
            if 'image' in service and 'swimlane' in service['image']:
                docker_image = DockerImage.parse(service['image'], self.config.args.dev)
                docker_image.version = version
                docker_compose_turbine.set(docker_image.full_name, 'services', name, 'image')
        docker_compose_main.override(docker_compose_turbine)
        docker_compose_main.save()

    @info_function_start_finish('Initializing Turbine database.')
    def initialize_turbine_database(self):
        # start Mongo
        main_docker_compose = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE)
        docker_compose_manager = DockerComposeManager(self.logger, main_docker_compose)
        docker_compose_manager.docker_compose_up(names.SW_MONGO)
        # access Mongo and make changes
        mongo_manager = MongoManager(self.config)
        user_name = self.db_configuration.get_sw_user_name()
        if not mongo_manager.user_exists(names.TURBINE_DATABASE, user_name):
            mongo_manager.create_user(names.TURBINE_DATABASE,
                                      user_name,
                                      self.db_configuration.get_sw_password(),
                                      create_database=True)
        # shut down service
        docker_compose_manager.docker_compose_down()

    @info_function_start_finish('Modifying env file.')
    def modify_env_file(self):
        # type: () -> None
        """
        Set the correct connection string to Mongo
        """
        db_key = 'DB_URL'
        turbine_env_file = path.join(self.secrets_dir, names.TURBINE_ENV_FILE)
        turbine_env_dict = read_dict(turbine_env_file)
        mongo_uri_dict = self.db_configuration.parse_mongo_uri(turbine_env_dict[db_key])
        mongo_uri_dict['username'] = self.db_configuration.get_sw_user_name()
        mongo_uri_dict['password'] = self.db_configuration.get_sw_password()
        turbine_env_dict[db_key] = self.db_configuration.get_mongo_uri(mongo_uri_dict)
        write_dict(turbine_env_file, turbine_env_dict)

    @info_function_start_finish('Copying turbine related files.')
    def copy_files(self):
        # type: () -> None
        """
        Copy turbine related files. Important for upgrade mostly.
        Currently install copies files using wild cards.
        """
        copy_if_not_exist(path.join(self.template_dir, names.DOCKER_COMPOSE_TURBINE_FILE),
                          path.join(self.template_dir, names.DOCKER_COMPOSE_TURBINE_FILE))
        copy_if_not_exist(path.join(self.template_dir, names.SECRETS_SUB_FOLDER, names.TURBINE_ENV_FILE),
                          path.join(self.secrets_dir, names.TURBINE_ENV_FILE))

    @debug_function_return
    def get_turbine_version_from_json(self):
        # type: () -> str
        """
        Returns turbine version.
        :return: version.
        """
        fp = open(path.join(self.template_dir, 'setup.json'))
        setup_json = load(fp)
        return setup_json.get('turbine_version')


def run(config):
    # type: (Configuration) -> None
    config.collect(turbine_enable_questions)
    if config.args.turbine_enable:
        EnableTurbine(config).run()



from swimlane_platform.lib import Automation, LogFileOptional
from swimlane_platform.lib.args_config_questions import AnswerRequiredValidator, \
    PathExistsValidator, \
    VersionValidator, \
    DnsValidator

automation_questions = [
    {
        'type': 'list',
        'name': 'automation',
        'message': 'Do you want to save your configuration selections for future use, or load previously saved '
                   'selections?',
        'choices': [
            {
                'name': 'No thanks',
                'value': Automation.Normal
            },
            {
                'name': 'Save',
                'value': Automation.Save
            },
            {
                'name': 'Load',
                'value': Automation.Load
            }
        ]
    },
    {
        'type': 'input',
        'name': 'automation_file',
        'message': 'Specify the full path to your configuration file.',
        'when': lambda a: 'automation' in a and a['automation'] != Automation.Normal
    }
]

logging_questions = [
    {
        'type': 'list',
        'message': 'What logging level do you want?',
        'name': 'verbose',
        'default': 0,
        'choices': [
            {
                'name': 'Info',
                'value': 0
            },
            {
                'name': 'Verbose',
                'value': 1
            },
            {
                'name': 'Debug',
                'value': 2
            }
        ]
    },
    {
        'type': 'input',
        'name': 'log',
        'message': 'If you would like program output written to a file, specify the path to the log '
                   'file (leave your answer blank if no log file is needed).',
        'validate': LogFileOptional
    }
]

dev_question = [
    {
        'type': 'confirm',
        'name': 'dev',
        'message': 'hidden question for cli overwrite only',
        'default': False,
        'when': lambda a: False
    }
]

web_ssl_questions = [
    {
        'type': 'confirm',
        'message': 'Do you want to use a self-signed certificate for SSL web connections? '
                   'If not, you must provide your own certificates.',
        'name': 'web_ssl_self_signed',
        'default': True
    },
    {
        'type': 'input',
        'name': 'web_ssl_certificate',
        'message': 'Specify the full path to your .crt certificate.',
        'when': lambda a: not a['web_ssl_self_signed'],
        'validate': PathExistsValidator
    },
    {
        'type': 'input',
        'name': 'web_ssl_key',
        'message': 'Specify the full path to your .key certificate.',
        'when': lambda a: not a['web_ssl_self_signed'],
        'validate': PathExistsValidator
    }]

ssl_questions = [
    {
        'type': 'input',
        'name': 'ssl_country',
        'message': 'Specify country.',
        'default': u'US',
        'validate': lambda a: a and 3 > len(a) > 1
    },
    {
        'type': 'input',
        'name': 'ssl_state',
        'message': 'Specify state.',
        'default': u'CO',
        'validate': lambda a: a and 3 > len(a) > 1
    },
    {
        'type': 'input',
        'name': 'ssl_location',
        'message': 'Specify location/city.',
        'default': u'Louisville',
        'validate': AnswerRequiredValidator
    },
    {
        'type': 'input',
        'name': 'ssl_company',
        'message': 'Specify company.',
        'default': u'Swimlane Inc',
        'validate': AnswerRequiredValidator
    },
    {
        'type': 'input',
        'name': 'ssl_application',
        'message': 'Specify application name.',
        'default': u'local.swimlane.io',
        'validate': AnswerRequiredValidator
    },
    {
        'type': 'input',
        'name': 'ssl_dns',
        'message': 'Specify at least one DNS name to store in the Subject Alternative Name (comma-separated).',
        'default': u'*.swimlane.io',
        'validate': DnsValidator
    }
]

database_encryption_key_question = {
    'type': 'input',
    'name': 'db_encryption_key',
    'message': 'Enter database encryption key?',
    'validate': AnswerRequiredValidator
}

mongo_admin_password_question = {
    'type': 'password',
    'name': 'mongo_admin_password',
    'message': 'MongoDB Admin user password?',
    'validate': AnswerRequiredValidator
}

mongo_sw_password_question = {
    'type': 'password',
    'name': 'mongo_sw_password',
    'message': 'MongoDB Swimlane user password?',
    'validate': AnswerRequiredValidator
}

turbine_enable_questions = [
    {
        'type': 'confirm',
        'message': 'Do you want to enable Turbine as your tasks execution engine.',
        'name': 'turbine_enable',
        'default': False,
        'when': lambda a: False
    },
    {
        'type': 'input',
        'message': 'What version of Turbine images you want to use?',
        'name': 'turbine_image_version',
        'when': lambda a: a.get('turbine_enable') and a.get('dev'),
        'validate': VersionValidator
    }
]

offline_questions = [
    {
        'type': 'confirm',
        'name': 'installer_is_offline',
        'message': 'Are you installing offline?'
    },
    {
        'type': 'input',
        'name': 'extracted_files_folder',
        'message': 'Please specify the path to the folder containing the extracted installer files.',
        'when': lambda a: 'installer_is_offline' in a and a['installer_is_offline'],
        'default': u'.',
        'validate': PathExistsValidator
    }
]



from swimlane_platform.lib.mongo_manager import MongoManager
from swimlane_platform.lib.docker_compose_manager import DockerComposeManager
from swimlane_platform.lib import names, env_manager
from swimlane_platform.lib.docker_compose_file_manager import DockerComposeFileManager
from swimlane_platform.lib.debug_decorators import info_function_start_finish, debug_function_args
from swimlane_platform.shared_steps import enable_turbine
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
from os import path
import semver


class UpgradeFrom1022To1030(UpgradeStep):
    FROM = semver.parse_version_info('10.2.2')  # type: semver.VersionInfo
    TO = semver.parse_version_info('10.3.0')  # type: semver.VersionInfo

    @info_function_start_finish('Upgrade From 10.2.2 To 10.3.0')
    def process(self):
        # type: () -> None
        enable_turbine.run(self.config)        
        self.upgrade_image_versions(names.INSTALL_DIR, self.config.args.dev)
        self.add_report_service(names.INSTALL_DIR)
        self.upgrade_compatibility_mode()
        self.add_env_vars(names.INSTALL_DIR)

    @debug_function_args
    def upgrade_image_versions(self, install_dir, dev):
        # type: (str, bool) -> None
        """
        Changes mongo image version to 4.4.2
        :param dev: If the images will be pulled from development repository.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        docker_compose = self.upgrade_standard_images(self.config.args.dev, names.INSTALL_DIR)
        docker_compose.set('mongo:4.4.2', 'services', names.SW_MONGO, 'image')
        docker_compose.save()

    @debug_function_args
    def add_report_service(self, install_dir):
        # type: (str, bool) -> None
        """
        Adds report service to docker-compose file
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        docker_compose = DockerComposeFileManager(self.logger, path.join(install_dir, names.DOCKER_COMPOSE_FILE))
        docker_compose.set(False, 'volumes', 'reports', 'external')
        docker_compose.set(['ALL'], 'services', names.SW_REPORTS, 'cap_drop')
        docker_compose.set(['CHOWN', 'KILL'], 'services', names.SW_REPORTS, 'cap_add')
        docker_compose.set(names.SW_REPORTS, 'services', names.SW_REPORTS, 'container_name')
        docker_compose.set(['no-new-privileges'], 'services', names.SW_REPORTS, 'security_opt')
        docker_compose.set('unless-stopped', 'services', names.SW_REPORTS, 'restart')
        docker_compose.set(['reports:/app'], 'services', names.SW_REPORTS, 'volumes')
        docker_compose.set({"internal_network": {"aliases": ['sw-reports']}}, 'services', names.SW_REPORTS, 'networks')

        docker_compose.save()

    @debug_function_args
    def upgrade_compatibility_mode(self):
        # type: () -> None
        """
        Issues database compatibility command.
        """
        docker_compose_file = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE)
        docker_compose_manager = DockerComposeManager(self.logger, docker_compose_file)
        docker_compose_manager.docker_compose_up(names.SW_MONGO)
        MongoManager(self.config).set_database_compatibility('4.4')
        docker_compose_manager.docker_compose_down()


    @debug_function_args
    def add_env_vars(self, install_dir):
        # type: (str, bool) -> None
        """
        Adds the new env variables to the api and tasks env files
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        # Modify env files.
        secrets_dir = path.join(install_dir, names.SECRETS_SUB_FOLDER)
        env_files = [names.API_ENV_FILE, names.TASKS_ENV_FILE]
        for file_path in (path.join(secrets_dir, file_name) for file_name in env_files):
            env_file = env_manager.read_dict(file_path)
            new_env_file = dict((key, value) for key, value in env_file.items())            
            new_env_file['SWIMLANE_FeatureManagement__DynamicOrchestration'] = 'false'
            new_env_file['SWIMLANE_FeatureManagement__ContentExchange'] = 'false'
            new_env_file['SWIMLANE_OpenSSL__MinProtocol'] = 'TLSv1.2'
            new_env_file['SWIMLANE_OpenSSL__CipherString'] = 'DEFAULT@SECLEVEL=2'
            new_env_file['SWIMLANE_ReportsService__BaseUrl'] = 'http://sw-reports:4000/'
            env_manager.write_dict(file_path, new_env_file)


from swimlane_platform.lib.args_config_questions import Configuration
from typing import Dict, List
from os import path
from swimlane_platform.lib import BaseWithLog, names, database_encryption_key_question, \
    mongo_admin_password_question, mongo_sw_password_question
from swimlane_platform.lib.env_manager import read_dict, write_dict
from swimlane_platform.lib.db_configuration import DbConfiguration
from swimlane_platform.lib.debug_decorators import info_function_start_finish, verbose_function_start_finish


class SetupSwimlaneEnvironment(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> SetupSwimlaneEnvironment
        super(SetupSwimlaneEnvironment, self).__init__(config)
        self.secrets_dir = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)
        self.db_configuration = DbConfiguration(self.config)

    @info_function_start_finish('Setup Swimlane environment.')
    def run(self):
        # type: () -> None
        """
        Main sequence of events.
        """
        self.setup_db_encryption_key()
        self.setup_passwords_in_env_file()

    @verbose_function_start_finish('Save db encryption key.')
    def setup_db_encryption_key(self):
        # type: () -> None
        """
        Writes entered db encryption key to the file used in docker-compose
        """
        with open(path.join(self.secrets_dir, names.DB_ENCRYPTION_KEY), 'w') as fs:
            fs.write(self.config.args.db_encryption_key)

    @verbose_function_start_finish('Configure env files.')
    def setup_passwords_in_env_file(self):
        # type: () -> None
        """
        Configure .env files
        """
        assert self.config.args.mongo_sw_password
        assert self.config.args.mongo_admin_password
        # Modify swimlane url in env file
        for env_file in [names.API_ENV_FILE, names.TASKS_ENV_FILE]:
            filename = path.join(self.secrets_dir, env_file)
            env_dict = read_dict(filename)
            for key in [names.DOT_NET_SWIMLANE_CONN_KEY, names.DOT_NET_HISTORY_CONN_KEY]:
                db_url = env_dict[key]
                mongo_settings = self.db_configuration.parse_mongo_uri(db_url)
                mongo_settings['password'] = self.config.args.mongo_sw_password
                env_dict[key] = self.db_configuration.get_mongo_uri(mongo_settings)
            write_dict(filename, env_dict)
            self.logger.verbose('{name} file was saved with new passwords.'.format(name=env_file))
        # Modify passwords in mongodb env
        mongo_env_filename = path.join(self.secrets_dir, names.MONGO_ENV_FILE)
        mongo_env_dict = read_dict(mongo_env_filename)
        mongo_env_dict[names.MONGO_ENV_ADMIN_PASSWORD] = self.config.args.mongo_admin_password
        mongo_env_dict[names.MONGO_ENV_SW_PASSWORD] = self.config.args.mongo_sw_password
        write_dict(mongo_env_filename, mongo_env_dict)
        self.logger.verbose('{name} file was saved with new passwords.'.format(name=names.MONGO_ENV_FILE))


def questions():
    # type: () -> List[Dict[str, str]]
    return [
        database_encryption_key_question,
        mongo_admin_password_question,
        mongo_sw_password_question
    ]


def run(config):
    # type: (Configuration) -> None
    """
    The script run method, that can be called by other script or from file run method.
    :param config: Configuration information collected by parent script.
    """
    config.collect(questions())
    SetupSwimlaneEnvironment(config).run()



# noinspection PyPep8Naming
class names:
    # general
    SWIMLANE_PREFIX = 'SWIMLANE_'
    APP_NAME = 'swimlane_platform'
    # database
    TURBINE_DATABASE = 'SwimlaneTurbine'
    DB_ENCRYPTION_KEY = 'database_encryption.key'
    DB_INIT_SCRIPT = 'init-mongodb-users.sh'
    # ssl related
    SSL_WEB_KEY = 'swimlane.key'
    SSL_WEB_CERTIFICATE = 'swimlane.crt'
    # docker files
    DOCKER_COMPOSE_FILE = 'docker-compose.yml'
    DOCKER_COMPOSE_TURBINE_FILE = 'docker-compose.turbine.yml'
    DOCKER_IMAGE_ARCHIVE_NAME = 'swimlane-images.tgz'
    TEMPLATE_REPOSITORY_TAG = '<sw_docker_image_repo_placeholder>'
    TEMPLATE_VERSION_TAG = '<sw_docker_image_tag_placeholder>'
    DEV_REPOSITORY = 'nexus.swimlane.io:5000/'
    DOCKER_COMPOSE_OVERRIDE_FILE = 'docker-compose.override.yml'
    DOCKER_COMPOSE_INSTALL_FILE = 'docker-compose.install.yml'
    # docker service names
    SW_API = 'sw_api'
    SW_WEB = 'sw_web'
    SW_NEO_J = 'sw_neo4j'
    SW_TASKS = 'sw_tasks'
    SW_MONGO = 'sw_mongo'
    SW_REPORTS = 'sw_reports'
    SW_TURBINE_ENGINE = 'sw_turbine_engine'
    # file and folder locations
    SECRETS_SUB_FOLDER = '.secrets/'
    DB_INIT_SUB_FOLDER = 'db-init'
    TEMPLATE_DIR = 'swimlane_template_dir'
    INSTALL_DIR = '/opt/swimlane'
    # environment files and variables
    TURBINE_ENV_FILE = '.turbine-env'
    API_ENV_FILE = '.api-env'
    TASKS_ENV_FILE = '.tasks-env'
    MONGO_ENV_FILE = '.mongo-env'
    MONGO_ENV_ADMIN_NAME = 'MONGO_INITDB_ROOT_USERNAME'
    MONGO_ENV_ADMIN_PASSWORD = 'MONGO_INITDB_ROOT_PASSWORD'
    MONGO_ENV_SW_PASSWORD = 'SW_MONGO_INITDB_SWIMLANE_PASSWORD'
    MONGO_ENV_SW_NAME = 'SW_MONGO_INITDB_SWIMLANE_USERNAME'
    DOT_NET_SWIMLANE_CONN_KEY = 'SWIMLANE_Data__Mongo__SwimlaneConnectionString'
    DOT_NET_HISTORY_CONN_KEY = 'SWIMLANE_Data__Mongo__HistoryConnectionString'
    DOT_NET_ADMIN_CONN_KEY = 'SWIMLANE_Data__Mongo__AdminConnectionString'

    def __init__(self):
        pass




from os import path
from swimlane_platform.lib import DockerComposeFileManager, DockerImage, names, \
    info_function_start_finish, debug_function_args
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
import semver


class UpgradeFrom801To900(UpgradeStep):
    FROM = semver.parse_version_info('8.0.1')  # type: semver.VersionInfo
    TO = semver.parse_version_info('9.0.0')  # type: semver.VersionInfo

    @info_function_start_finish('Upgrade From 8.0.1 To 9.0.0')
    def process(self):
        # type: () -> None
        self.upgrade_image_versions(names.INSTALL_DIR, self.config.args.dev)

    @debug_function_args
    def upgrade_image_versions(self, install_dir, dev):
        # type: (str, bool) -> None
        """
        Changes image versions to the new ones
        :param dev: If the images will be pulled from development repository.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        docker_compose = self.upgrade_standard_images(dev, install_dir)
        docker_compose.set('mongo:4.0.9', 'services', names.SW_MONGO, 'image')
        docker_compose.save()




from os import path, remove
from semver import VersionInfo
from swimlane_platform.shared_steps import enable_mongo_ssl
from swimlane_platform.lib.args_config_questions import Configuration
from swimlane_platform.lib import DockerComposeFileManager, env_manager, names, \
    info_function_start_finish, debug_function_args, DockerManager
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
import shutil
from swimlane_platform.add_file_encryption_key import run as run_add_file_encryption_key
import semver


class UpgradeFrom701To800(UpgradeStep):
    FROM = semver.parse_version_info('7.0.1')  # type: VersionInfo
    TO = semver.parse_version_info('8.0.0')  # type: VersionInfo

    @info_function_start_finish('Upgrade From 7.0.1 To 8.0.0')
    def process(self):
        # type: () -> None
        self.remove_neo4j(names.INSTALL_DIR)
        enable_mongo_ssl.run(self.config)
        self.add_swimlane_prefix(names.INSTALL_DIR)
        self.upgrade_image_versions(names.INSTALL_DIR, self.config.args.dev)
        self.add_encryption_keys(self.config)
        self.add_default_ha(names.INSTALL_DIR)
        self.add_dashed_aliases(names.INSTALL_DIR)
        self.remove_pip_volume()
        self.change_restart_policy(names.INSTALL_DIR)
        self.add_pfx_password_template_to_env_files(names.INSTALL_DIR)

    @debug_function_args
    def remove_neo4j(self, install_dir):
        # type: (str) -> None
        """
        Removes all mentions of neo4j from infrastructure.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        # Modify env files.
        secrets_dir = path.join(install_dir, names.SECRETS_SUB_FOLDER)
        env_files = [names.API_ENV_FILE, names.TASKS_ENV_FILE]
        for file_path in (path.join(secrets_dir, file_name) for file_name in env_files):
            env_file = env_manager.read_dict(file_path)
            new_env_file = dict((key, value) for key, value in env_file.items() if 'Neo4j' not in key)
            wait_for = env_file.get('WAIT_FOR').split(',')
            new_wait_for = [service for service in wait_for if names.SW_NEO_J not in service]
            new_env_file['WAIT_FOR'] = ','.join(new_wait_for)
            env_manager.write_dict(file_path, new_env_file)
        # Remove unneeded env file
        neo4j_env_file = path.join(secrets_dir, '.neo4j-env')
        if path.exists(neo4j_env_file):
            remove(neo4j_env_file)
        # Modify docker-compose file.
        docker_compose = DockerComposeFileManager(self.logger, path.join(install_dir, names.DOCKER_COMPOSE_FILE))
        docker_compose.get('services').pop(names.SW_NEO_J, None)
        docker_compose_override = DockerComposeFileManager(self.logger,
                                                           path.join(install_dir, names.DOCKER_COMPOSE_OVERRIDE_FILE))
        docker_compose_override.get('services').pop(names.SW_NEO_J, None)
        docker_compose_override.save()
        for service_name in [names.SW_API, names.SW_TASKS]:
            docker_compose.subtract_from_list(names.SW_NEO_J, 'services', service_name, 'depends_on')
        docker_compose.get('volumes').pop('neo4jdb', None)
        docker_compose.save()

    @debug_function_args
    def add_swimlane_prefix(self, install_dir):
        # type: (str) -> None
        """
        Adds SWIMLANE_ prefix to dotnet config variables. Made for better separation and management of them.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        swimlane_prefix = 'SWIMLANE_'

        def append_prefix(key):
            # type: (str) -> str
            return swimlane_prefix + key if '__' in key and not key.startswith(swimlane_prefix) else key

        secrets_dir = path.join(install_dir, names.SECRETS_SUB_FOLDER)
        env_files = [names.API_ENV_FILE, names.TASKS_ENV_FILE]
        for file_path in (path.join(secrets_dir, file_name) for file_name in env_files):
            env_file = env_manager.read_dict(file_path)
            new_env_file = dict((append_prefix(key), value) for key, value in env_file.items())
            env_manager.write_dict(file_path, new_env_file)

    @debug_function_args
    def upgrade_image_versions(self, install_dir, dev):
        # type: (str, bool) -> None
        """
        Changes image versions to the new ones
        :param dev: If the images will be pulled from development repository.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        docker_compose = self.upgrade_standard_images(dev, install_dir)
        docker_compose.set('mongo:4.0.5', 'services', names.SW_MONGO, 'image')
        docker_compose.save()

    @debug_function_args
    def add_encryption_keys(self, config):
        # type: (Configuration) -> None
        """
        Implements changes to the secrets.
        :param config: Application arguments.
        """
        # Rename .txt to .key
        secrets_dir = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)
        old_database_encryption_key = path.join(secrets_dir, 'database_encryption_key.txt')
        if path.exists(old_database_encryption_key):
            new_database_encryption_key = path.join(secrets_dir, 'database_encryption.key')
            shutil.move(old_database_encryption_key, new_database_encryption_key)
            docker_compose = DockerComposeFileManager(self.logger,
                                                      path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE))
            docker_compose.set(new_database_encryption_key, 'secrets', 'database_encryption_key', 'file')
            docker_compose.save()
        run_add_file_encryption_key(config)

    @debug_function_args
    def add_default_ha(self, install_dir):
        secrets_dir = path.join(install_dir, names.SECRETS_SUB_FOLDER)
        env_files = [names.API_ENV_FILE, names.TASKS_ENV_FILE]
        for file_path in (path.join(secrets_dir, file_name) for file_name in env_files):
            env_file = env_manager.read_dict(file_path)
            if 'SWIMLANE_EnableHA' not in env_file:
                env_file['SWIMLANE_EnableHA'] = 'false'
            env_manager.write_dict(file_path, env_file)

    @debug_function_args
    def add_dashed_aliases(self, install_dir):
        docker_compose = DockerComposeFileManager(self.logger, path.join(install_dir, names.DOCKER_COMPOSE_FILE))
        for service in [names.SW_API, names.SW_WEB]:
            alias_path = ['services', service, 'networks', 'internal_network', 'aliases']
            docker_compose.append_or_create_list(service.replace('_', '-'), *alias_path)
        docker_compose.save()

    @debug_function_args
    def remove_pip_volume(self):
        """
        Removes the pip volume so it can be rebuilt by the application. Preinstalled packages were removed
        from the volume in version 8.0.0.
        """
        docker_manager = DockerManager(self.logger)
        docker_manager.volume_remove('pip')

    @debug_function_args
    def change_restart_policy(self, install_dir):
        """
        The restart policy for swimlane containers was changed from "on-failure:5" to "unless-stopped" in version 8.0.0
        """
        docker_compose = DockerComposeFileManager(self.logger, path.join(install_dir, names.DOCKER_COMPOSE_FILE))
        for service in [names.SW_API, names.SW_WEB, names.SW_TASKS]:
            alias_path = ['services', service, 'restart']
            docker_compose.set('unless-stopped', *alias_path)
        docker_compose.save()

    @debug_function_args
    def add_pfx_password_template_to_env_files(self, install_dir):
        """
        Adds the commented out SWIMLANE_PfxPassword env variable to the API and Tasks env files
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        secrets_dir = path.join(install_dir, names.SECRETS_SUB_FOLDER)
        env_files = [names.API_ENV_FILE, names.TASKS_ENV_FILE]
        key_name = '#{}_PfxPassword'.format(names.SWIMLANE_PREFIX)
        for file_path in (path.join(secrets_dir, file_name) for file_name in env_files):
            env_file = env_manager.read_dict(file_path)
            if any(key in env_file for key in (key_name, key_name[1:])):
                return
            
            env_file[key_name] = '<password-placeholder>'
            env_manager.write_dict(file_path, env_file)



class Actions:

    def __init__(self):
        pass

    Install = 'Install'
    Upgrade = 'Upgrade'
    Backup = 'Backup'
    Restore = 'Restore'


class Automation:

    def __init__(self):
        pass

    Normal = 'normal'
    Save = 'save'
    Load = 'load'



from swimlane_platform.lib import names, env_manager
from swimlane_platform.lib.debug_decorators import info_function_start_finish, debug_function_args
from swimlane_platform.shared_steps import enable_turbine
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
from os import path
import semver


class UpgradeFrom1030To1040rc2(UpgradeStep):
    FROM = semver.parse_version_info('10.3.0')  # type: semver.VersionInfo
    TO = semver.parse_version_info('10.4.0-rc2')  # type: semver.VersionInfo

    @info_function_start_finish('Upgrade From 10.3.0 To 10.4.0-rc2')
    def process(self):
        # type: () -> None
        enable_turbine.run(self.config)
        self.upgrade_image_versions(names.INSTALL_DIR, self.config.args.dev)
        self.update_env_vars(names.INSTALL_DIR)

    @debug_function_args
    def upgrade_image_versions(self, install_dir, dev):
        # type: (str, bool) -> None
        """
        Updates image versions
        :param dev: If the images will be pulled from development repository.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        docker_compose = self.upgrade_standard_images(self.config.args.dev, names.INSTALL_DIR)
        docker_compose.save()

    @debug_function_args
    def update_env_vars(self, install_dir):
        # type: (str, bool) -> None
        """
        Updates environment variables in .env files
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        # Modify env files.
        secrets_dir = path.join(install_dir, names.SECRETS_SUB_FOLDER)
        env_files = [names.API_ENV_FILE, names.TASKS_ENV_FILE]
        for file_path in (path.join(secrets_dir, file_name) for file_name in env_files):
            env_file = env_manager.read_dict(file_path)
            new_env_file = dict((key, value) for key, value in env_file.items())
            new_env_file.pop('SWIMLANE_FeatureManagement__ContentExchange', None)
            env_manager.write_dict(file_path, new_env_file)


from os import path
from semver import VersionInfo
from swimlane_platform.lib import DockerComposeFileManager, DockerImage, env_manager, names, \
    info_function_start_finish, debug_function_args
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
import semver


class UpgradeFrom800To801(UpgradeStep):
    FROM = semver.parse_version_info('8.0.0')  # type: VersionInfo
    TO = semver.parse_version_info('8.0.1')  # type: VersionInfo

    @info_function_start_finish('Upgrade From 8.0.0 To 8.0.1')
    def process(self):
        # type: () -> None
        self.upgrade_image_versions(names.INSTALL_DIR, self.config.args.dev)
        self.add_requests_ca_bundle_to_env_files(names.INSTALL_DIR)

    @debug_function_args
    def upgrade_image_versions(self, install_dir, dev):
        # type: (str, bool) -> None
        """
        Changes image versions to the new ones
        :param dev: If the images will be pulled from development repository.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        self.upgrade_standard_images(dev, install_dir).save()

    @debug_function_args
    def add_requests_ca_bundle_to_env_files(self, install_dir):
        """
        Adds the REQUESTS_CA_BUNDLE env variable to the API and Tasks env files. This env var was added
        in version 8.0.0 and is always on.
        :param install_dir: Root folder for installation. Where docker-compose resides.
        """
        secrets_dir = path.join(install_dir, names.SECRETS_SUB_FOLDER)
        env_files = [names.API_ENV_FILE, names.TASKS_ENV_FILE]
        key_name = 'REQUESTS_CA_BUNDLE'
        for file_path in (path.join(secrets_dir, file_name) for file_name in env_files):
            env_file = env_manager.read_dict(file_path)
            if key_name not in env_file:
                env_file[key_name] = '/etc/ssl/certs'
                env_manager.write_dict(file_path, env_file)



from __future__ import print_function
from os import path, walk, unlink, makedirs
import shutil
from datetime import datetime
import docker
from typing import Union
from swimlane_platform.lib.db_configuration import DbConfiguration
from swimlane_platform.lib import debug_function_return, debug_function_args, DockerManager, names, \
    ValidationException, parse_env_variables_to_dict, Configuration
from swimlane_platform.lib.base_with_log import BaseWithLog


class MongoConnection:
    host = None
    port = None
    user_name = None
    password = None
    ssl = False
    pem_file = None

    def __init__(self, uri=None, config=None, pem_file=None):
        # type: (Union[str, None], Union[Configuration, None], Union[str, None]) -> None
        if uri and config:
            raise ValidationException("MongoCollection accepts either uri or configuration, not both.")
        elif uri:
            self.from_uri(uri, pem_file)
        elif config:
            self.from_configuration(config, pem_file)
        else:
            raise ValidationException("MongoCollection accepts either uri or configuration, none were provided.")

    def from_uri(self, uri, pem_file):
        # type: (str, str) -> None
        db_configuration = DbConfiguration(Configuration())
        uri = db_configuration.parse_mongo_uri(uri)
        host, port = uri.get('nodelist')[0]
        self.host = host
        self.port = port
        self.user_name = uri.get('username')
        self.password = uri.get('password')
        self.ssl = 'options' in uri and uri['options'].get('ssl')
        self.pem_file = pem_file

    def from_configuration(self, config, pem_file):
        # type: (Configuration, object) -> None
        self.host = config.args.mongo_host
        self.port = config.args.mongo_port
        self.user_name = config.args.mongo_user_name
        self.password = config.args.mongo_password
        self.ssl = bool(pem_file)
        self.pem_file = pem_file


class BackupRestoreBase(BaseWithLog):
    BACKUP_IMAGE = None
    DB_SWIMLANE = 'Swimlane'
    DB_SWIMLANE_HISTORY = 'SwimlaneHistory'
    BACKUP_FOLDER = '/backup'
    CERT_FOLDER = '/certs'

    def __init__(self, config):
        # type: (Configuration) -> None
        super(BackupRestoreBase, self).__init__(config)
        self.client = docker.from_env()
        self.docker_manager = DockerManager(self.logger)
        backup_image_name = 'swimlane/swimlane-backup:latest'
        self.BACKUP_IMAGE = 'nexus.swimlane.io:5000/' + backup_image_name if self.config.args.dev else backup_image_name

    @debug_function_return
    @debug_function_args
    def clear_directory(self, folder_path):
        # type: (str) -> None
        """
        Removes all files from the directory.
        :param folder_path: Directory
        """
        for root, dirs, files in walk(folder_path):
            for f in files:
                unlink(path.join(root, f))
            for d in dirs:
                shutil.rmtree(path.join(root, d))

    @debug_function_return
    @debug_function_args
    def get_archive_name(self, backup_dir, archive=None):
        # type: (str, str) -> (str, str)
        """
        Returns file name for new archive.
        :param archive: Archive name if known (for run)
        :param backup_dir: Location on host to put backups.
        :return: File name
        """
        name = archive if archive else datetime.now().strftime("%Y%m%d-%H%M")
        remote = '/'.join([self.BACKUP_FOLDER, name])
        local = path.join(backup_dir, name)
        if not path.exists(local):
            makedirs(local)
        return local, remote

    @debug_function_return
    def get_certificate_location(self):
        # type: () -> Union[str, None]
        mongo_container = self.docker_manager.container_get(names.SW_MONGO)
        if not mongo_container:
            raise ValidationException('Cannot find MongoDb container on the system.')
        mongo_container_attr = mongo_container.attrs
        cmd = mongo_container_attr.get('Config', {}).get('Cmd')
        pem_key_file = '--tlsCertificateKeyFile'
        if not cmd or pem_key_file not in cmd:
            return None
        pem_switch = cmd.index(pem_key_file)
        if len(cmd) < pem_switch + 2:
            return None
        pem_secret_location = cmd[pem_switch + 1]
        for mount in mongo_container_attr.get('Mounts', []):
            if mount.get('Destination') == pem_secret_location:
                local_file_path = mount.get('Source')
                if local_file_path:
                    local_file_path = path.abspath(local_file_path)
                    if path.exists(local_file_path):
                        return local_file_path
                    else:
                        self.logger.warn('MongoDB pem file was found in container, but not on the file system.')
        return None

    @debug_function_args
    def validate_backup_exists(self, full_archive_path, *backups):
        # type: (str, str) -> None
        """
        Validates if individual backups are actually present in the folder.
        :param full_archive_path: Backup folder.
        :param backups: Names of backups expected to be in the folder.
        """
        for name in backups:
            if not path.exists(path.join(full_archive_path, name)):
                raise ValidationException("{name} backup is missing".format(name=name))

    def report_archive_name(self, archive_name):
        # type: (str) -> None
        self.logger.info(
            'Backup completed successfully. Archive name: {archive_name}'.format(archive_name=archive_name))

    @debug_function_return
    @debug_function_args
    def get_mongo_connection(self):
        # type: () -> MongoConnection
        """
        Acquires Mongo Connection either form API container or Questions
        :return: Mongo Connection
        """
        env = parse_env_variables_to_dict(self.docker_manager.container_get_dotnet_env(names.SW_API))
        pem_file = self.get_certificate_location()
        if not pem_file:
            questions = [
                {
                    'type': 'input',
                    'name': 'mongo_certificate',
                    'message': 'Location of Mongo certificate',
                    'validate': lambda a: path.exists(a)
                }
            ]
            self.config.collect(questions)
            pem_file = self.config.args.mongo_certificate
        if names.DOT_NET_SWIMLANE_CONN_KEY not in env:
            questions = [
                {
                    'type': 'input',
                    'name': 'mongo_host',
                    'message': 'Host name of mongo server',
                    'default': names.SW_MONGO
                },
                {
                    'type': 'input',
                    'name': 'mongo_port',
                    'message': 'Port of mongo server',
                    'default': 27017
                },
                {
                    'type': 'input',
                    'name': 'mongo_user_name',
                    'message': 'Mongo user name'
                },
                {
                    'type': 'input',
                    'name': 'mongo_password',
                    'message': 'Mongo password'
                }
            ]
            self.config.collect(questions)
            return MongoConnection(config=self.config, pem_file=pem_file)
        else:
            return MongoConnection(uri=env[names.DOT_NET_SWIMLANE_CONN_KEY], pem_file=pem_file)



from os import path
from typing import Union, List, Callable, Dict, Any, Tuple
from swimlane_platform.lib.args_config_questions import Configuration
from swimlane_platform.lib import mongo_admin_password_question, mongo_sw_password_question
from swimlane_platform.lib.expected_exceptions import ValidationException
from swimlane_platform.lib.names import names
from swimlane_platform.lib.env_manager import read_dict
from swimlane_platform.lib.base_with_log import BaseWithLog
from swimlane_platform.lib.debug_decorators import debug_function_return, debug_function_args
from future.standard_library import install_aliases
install_aliases()
from urllib.parse import quote
from pymongo import uri_parser


class DbConfiguration(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> None
        super(DbConfiguration, self).__init__(config)
        self.secrets_dir = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)

    @debug_function_return
    @debug_function_args
    def get_mongo_uri(self, mongo_settings):
        # type: (Union[None,Dict[str, Any]]) -> str
        """
        Returns formatted MongoDb Uri.
        :param mongo_settings: MongoDb settings.
        :return: MongoDb uri string.
        """
        url = 'mongodb://'
        user_name = mongo_settings.get('username', '')
        password = quote(mongo_settings.get('password', ''))
        if user_name or password:
            url += '{user}:{password}@'.format(user=user_name, password=password)
        url += ','.join(['{host}:{port}'.format(host=node[0], port=node[1]) for node in mongo_settings['nodelist']])
        if mongo_settings.get('database'):
            url += '/{database}'.format(database=mongo_settings['database'])
        if mongo_settings.get('options'):
            url += '?'
            items = ['{key}={value}'.format(key=key, value=value) for key, value in mongo_settings['options'].items()]
            url += '&'.join(items)
        return url

    @debug_function_return
    @debug_function_args
    def parse_mongo_uri(self, uri):
        # type: (str) -> Dict[str, Union[str, List[Tuple[str, str]], Dict[str, str]]]
        return uri_parser.parse_uri(uri, validate=False)

    @debug_function_return
    def get_sw_user_name(self):
        # type: () -> str
        """
        Gets swimlane user name.
        :return: swimlane user name.
        :exception: ValidationException if cannot find user name.
        """
        return self._chain('Cannot find swimlane user name.',
                           lambda: self._inspect_mongo_env(names.MONGO_ENV_SW_NAME),
                           lambda: self._inspect_api_env_swimlane_connection('username')
                           )

    @debug_function_return
    def get_sw_password(self):
        # type: () -> str
        """
        Gets swimlane user password.
        :return: swimlane user password.
        :exception: ValidationException if cannot find password.
        """
        return self._chain('Cannot find swimlane user password.',
                           lambda: self._inspect_answered_questions(mongo_sw_password_question),
                           lambda: self._inspect_mongo_env(names.MONGO_ENV_SW_PASSWORD),
                           lambda: self._inspect_api_env_swimlane_connection('password')
                           )

    @debug_function_return
    def get_admin_user_name(self):
        # type: () -> str
        """
        Gets admin user name.
        :return: admin user name.
        :exception: ValidationException if cannot find user name.
        """
        return self._chain('Cannot find admin user name.',
                           lambda: self._inspect_mongo_env(names.MONGO_ENV_ADMIN_NAME))

    @debug_function_return
    def get_admin_password(self):
        # type: () -> str
        """
        Gets admin password.
        :return: admin password.
        :exception: ValidationException if cannot find password.
        """
        return self._chain('Cannot find admin user password.',
                           lambda: self._inspect_answered_questions(mongo_admin_password_question),
                           lambda: self._inspect_mongo_env(names.MONGO_ENV_ADMIN_PASSWORD)
                           )

    @debug_function_return
    @debug_function_args
    def _inspect_api_env_swimlane_connection(self, key):
        # type: (str) -> Union[str, None]
        """
        Returns value for the key in Swimlane connection string. Taken from .api-env
        :param key: Key in connection string. i.e. username, password, database etc
        :return: None if not present or the value found.
        """
        api_env = read_dict(path.join(self.secrets_dir, names.API_ENV_FILE))
        swimlane_connection = api_env.get(names.DOT_NET_SWIMLANE_CONN_KEY)
        if swimlane_connection:
            uri = self.parse_mongo_uri(swimlane_connection)
            return uri.get(key)
        return None

    @debug_function_return
    @debug_function_args
    def _inspect_mongo_env(self, key):
        # type: (str) -> Union[None, str]
        """
        Returns value found in .mongo-env for a key.
        :param key: Key, i.e swimlane user, admin password. Correct key names are in names
        :return: None if not found or value.
        """
        mongo_env = read_dict(path.join(self.secrets_dir, names.MONGO_ENV_FILE))
        return mongo_env.get(key)

    @debug_function_return
    @debug_function_args
    def _inspect_answered_questions(self, question):
        # type: (Dict[str, Any]) -> Union[None, str]
        """
        Returns value if the question has been answered.
        :param question: question to search for in answered args object.
        :return: None if not found or value.
        """
        question_name = question.get('name')
        return self.config.get(question_name) if question_name else None

    @debug_function_return
    @debug_function_args
    def _chain(self, error, *functions):
        # type: (str, List[Callable]) -> str
        """
        Evaluate and returns first value from submitted functions.
        :param error: Exception to throw if values are not found.
        :param functions: Lambda functions to evaluate to get value. Done for performance reasons.
        :return: value or raise specified error.
        """
        for f in functions:
            value = f()
            if value:
                return value
        raise ValidationException(error)



from swimlane_platform.lib.names import names
from swimlane_platform.lib.debug_decorators import debug_function_args


class DockerImage:

    def __init__(self, repository=names.DEV_REPOSITORY, image_name=None, version=None, is_dev=False):
        # type: (str, str, str, bool) -> DockerImage
        """
        Constructor
        :param repository: Repository name.
        :param image_name: Image name.
        :param version: Version.
        :param is_dev: True if running in development environment.
        """
        self.repository = repository
        self.image_name = image_name
        self.version = version
        self.is_dev = is_dev

    @debug_function_args
    def _parse(self, tag):
        # type: (str) -> None
        """
        Parses full tag to object properties.
        :param tag: The full tag.
        """
        tag = tag.replace(self.repository, '')
        self.image_name, self.version = tag.rsplit(':', 1)

    @staticmethod
    def parse(tag, is_dev):
        # type: (str, bool) -> DockerImage
        """
        Parses swimlane image tag and returns DockerImage
        :param tag: The full image value from docker-compose
        :param is_dev: If image is pulled from development repository.
        :return: DockerImage
        """
        docker_image = DockerImage(is_dev=is_dev)
        docker_image._parse(tag)
        return docker_image

    @property
    def archive_name(self):
        # type: () -> str
        """
        Returns the archive image for docker image.
        :return: Archive image name.
        """
        return '{image}.tgz'.format(image=self.image_name.split('/')[-1])

    @property
    def full_name(self):
        # type: () -> str
        """
        Combines docker image tag parts together.
        :return: Full image tag.
        """
        return '{repository}{image}:{version}' \
            .format(repository=self.repository if self.is_dev else '', image=self.image_name, version=self.version)



from os import path
from swimlane_platform.backup.backup_restore_base import BackupRestoreBase
from swimlane_platform.lib import debug_function_args, names, PathExistsValidator, \
    AnswerRequiredValidator, info_function_start_finish


class Restore(BackupRestoreBase):

    @info_function_start_finish('Swimlane restore.')
    def run(self):
        """
        Main method.
        """
        host_archive, docker_archive = self.get_archive_name(self.config.args.backup_dir, self.config.args.archive)
        # noinspection PyBroadException
        try:
            self.docker_manager.containers_exists_validate(names.SW_TASKS, names.SW_WEB, names.SW_API, names.SW_MONGO)
            self.validate_backup_exists(host_archive,
                                        '{db}.gz'.format(db=self.DB_SWIMLANE),
                                        '{db}.gz'.format(db=self.DB_SWIMLANE_HISTORY))
            self.docker_manager.containers_run(self.docker_manager.container_stop, names.SW_WEB, names.SW_TASKS,
                                               names.SW_API)
            self.restore_mongo(docker_archive, self.DB_SWIMLANE)
            self.restore_mongo(docker_archive, self.DB_SWIMLANE_HISTORY, True)
            self.remove_pip_volume()
        finally:
            self.docker_manager.containers_run(self.docker_manager.container_start, names.SW_API, names.SW_TASKS,
                                               names.SW_WEB)

    @debug_function_args
    def restore_mongo(self, archive, database, auto_remove=False):
        # type: (str, str, bool) -> None
        """
        Restore MongoDb database.
        :rtype: None
        :param archive:
        :param database:
        :param auto_remove:
        """
        self.logger.info("Started MongoDb {db} run.".format(db=database))
        mongo_connection = self.get_mongo_connection()
        folder, cert_name = path.split(mongo_connection.pem_file)
        drop_command = 'mongo' \
                       ' {db}' \
                       ' --host {host}' \
                       ' --username {mongo_user}' \
                       ' --password {mongo_password}' \
                       ' --eval "db.dropDatabase();"' \
                       ' --ssl' \
                       ' --sslPEMKeyFile {cert_folder}/{cert_name}' \
                       ' --sslAllowInvalidCertificates'.format(db=database,
                                                               host=mongo_connection.host,
                                                               mongo_user=mongo_connection.user_name,
                                                               mongo_password=mongo_connection.password,
                                                               cert_name=cert_name,
                                                               cert_folder=self.CERT_FOLDER)
        restore_command = 'mongorestore' \
                          ' --host {host}' \
                          ' --gzip' \
                          ' --ssl' \
                          ' --drop' \
                          ' --db {db}' \
                          ' --archive={archive}/{db}.gz' \
                          ' --username {mongo_user}' \
                          ' --password {mongo_password}'\
                          ' --ssl' \
                          ' --sslPEMKeyFile {cert_folder}/{cert_name}' \
                          ' --sslAllowInvalidCertificates'.format(host=mongo_connection.host,
                                                                  db=database,
                                                                  mongo_user=mongo_connection.user_name,
                                                                  mongo_password=mongo_connection.password,
                                                                  archive=archive,
                                                                  cert_name=cert_name,
                                                                  cert_folder=self.CERT_FOLDER)
        volumes = {
            self.config.args.backup_dir: {'bind': self.BACKUP_FOLDER, 'mode': 'Z'},
            folder: {'bind': self.CERT_FOLDER, 'mode': 'rw'}
        }
        command = ['/bin/bash', '-c', '{drop}; {restore}'.format(drop=drop_command, restore=restore_command)]
        network = self.docker_manager.container_get_network_name(names.SW_MONGO)
        logs = self.client.containers.run(self.BACKUP_IMAGE,
                                          command=command,
                                          stderr=True,
                                          stream=True,
                                          auto_remove=auto_remove,
                                          volumes=volumes,
                                          network=network)
        for log in logs:
            self.logger.verbose(log)
        self.logger.info("Finished MongoDb {db} restore.".format(db=database))

    @info_function_start_finish()
    def remove_pip_volume(self):
        # type: () -> None
        """
        Remove pip volume.
        """
        self.docker_manager.volume_remove('pip')


def run(config):
    questions = [
        {
            'type': 'input',
            'name': 'backup_dir',
            'message': 'Folder where backups are stored',
            'validate': PathExistsValidator
        },
        {
            'type': 'input',
            'name': 'archive',
            'message': 'Name of the archive to run from',
            'validate': AnswerRequiredValidator
        }
    ]
    config.collect(questions)
    Restore(config).run()



from swimlane_platform.lib.expected_exceptions import ExpectedException, ValidationException
from swimlane_platform.lib.debug_decorators import debug_function_return, debug_function_args, \
    info_function_start_finish
from swimlane_platform.lib.docker_compose_manager import DockerComposeManager
from swimlane_platform.lib.docker_compose_file_manager import DockerComposeFileManager
from swimlane_platform.lib.docker_manager import DockerManager
from swimlane_platform.lib.names import names
from swimlane_platform.lib.args_config_questions import Configuration, AnswerRequiredValidator, \
    PathExistsValidator, Arguments, LogFileOptional, VersionValidator, ContainingDirectoryExistsValidator
from swimlane_platform.lib.env_manager import *
from swimlane_platform.lib.logger import SplitStreamLogger
from swimlane_platform.lib.docker_image import DockerImage
from swimlane_platform.lib.models import Actions, Automation
from swimlane_platform.lib.questions_groups import *
from swimlane_platform.lib.base_with_log import BaseWithLog
from swimlane_platform.lib.functions import *
from swimlane_platform.lib.open_ssl_manager import *
from swimlane_platform.lib.db_configuration import DbConfiguration
from swimlane_platform.lib.mongo_manager import MongoManager



from swimlane_platform.lib import names
from swimlane_platform.lib.base_with_log import BaseWithLog
from os import path
from swimlane_platform.lib.args_config_questions import Configuration
from stat import *
from swimlane_platform.lib.os_manager import OsManager
from swimlane_platform.lib.constants import SWIMLANE_UID_GID
from swimlane_platform.lib.debug_decorators import info_function_start_finish

class SetupPermissions(BaseWithLog):

    SECRETS_DIR_PERMS = S_IRWXU
    SECRETS_FILE_PERMS = S_IREAD | S_IWRITE
    SSL_WEB_CERTIFICATE_PERMS = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH
    API_TASKS_UID_GID = SWIMLANE_UID_GID
    WEB_UID_GID = SWIMLANE_UID_GID

    def __init__(self, config):
        # type: (Configuration) -> SetupPermissions
        super(SetupPermissions, self).__init__(config)
        self.os_manager = OsManager(self.logger)
        self.secrets_dir = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)

    @info_function_start_finish('Setup Permissions.')
    def run(self):
        # type: () -> None
        """
        Main sequence of events.
        """
        self.check_and_set_perms(self.secrets_dir, self.SECRETS_DIR_PERMS)
        self.check_and_set_perms(path.join(self.secrets_dir, names.SSL_WEB_CERTIFICATE), self.SSL_WEB_CERTIFICATE_PERMS)
        self.check_and_set_perms(path.join(self.secrets_dir, names.TASKS_ENV_FILE), self.SECRETS_FILE_PERMS)
        self.check_and_set_perms(path.join(self.secrets_dir, names.MONGO_ENV_FILE), self.SECRETS_FILE_PERMS)
        self.check_and_set_perms(path.join(self.secrets_dir, names.API_ENV_FILE), self.SECRETS_FILE_PERMS)
        self.check_and_set_owner(path.join(self.secrets_dir, names.SSL_WEB_KEY), self.WEB_UID_GID, self.WEB_UID_GID)
        self.check_and_set_perms(path.join(self.secrets_dir, names.SSL_WEB_KEY), self.SECRETS_FILE_PERMS)
        self.check_and_set_owner(path.join(self.secrets_dir, names.DB_ENCRYPTION_KEY), self.API_TASKS_UID_GID,
                                 self.API_TASKS_UID_GID)
        self.check_and_set_perms(path.join(self.secrets_dir, names.DB_ENCRYPTION_KEY), self.SECRETS_FILE_PERMS)

    # noinspection PyBroadException
    def check_and_set_perms(self, file_system_path, permission):
        # type: (str, int) -> None
        """
        Verifies permission and sets it if it's different.
        :param file_system_path: File or folder path.
        :param permission: Octal number for permission.
        """
        if self.os_manager.get_permissions_linux(file_system_path) != permission:
            try:
                self.os_manager.set_permissions_linux(file_system_path, permission)
                self.logger.verbose('Changed permission on {name} to {perms}'
                                    .format(name=file_system_path, perms=permission))
            except Exception:
                self.logger.warn('Could not change permissions on {file}.'.format(file=file_system_path))
                self.logger.warn('You should do it manually "chmod {file} {perms}"'
                                 .format(file=file_system_path, perms=permission))

    # noinspection PyBroadException
    def check_and_set_owner(self, file_system_path, uid, gid):
        # type: (str, int, int) -> None
        """
        Verifies owner on file system object and sets it to correct one if wrong
        :param file_system_path: File or folder path.
        :param uid: User id.
        :param gid: Group id.
        """
        _uid, _gid = self.os_manager.get_owner_linux(file_system_path)
        if uid != _uid or gid != _gid:
            try:
                self.os_manager.set_owner_linux(file_system_path, uid, gid)
                self.logger.verbose('Changed owner on {name} to {uid}:{gid}'
                                    .format(name=file_system_path, uid=uid, gid=gid))
            except Exception:
                self.logger.warn('Could not change owner on {file}.'.format(file=file_system_path))
                self.logger.warn('You should do it manually "sudo chown {uid}:{gid} {file}"'
                                 .format(file=file_system_path, uid=uid, gid=gid))


def run(config):
    # type: (Configuration) -> None
    """
    The script run method, that can be called by other script or from file run method.
    :param config: Configuration information collected by parent script.
    """
    SetupPermissions(config).run()



from typing import List, Type
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
from swimlane_platform.upgrade_steps.upgrade_from_700_701 import UpgradeFrom700To701
from swimlane_platform.upgrade_steps.upgrade_from_701_800 import UpgradeFrom701To800
from swimlane_platform.upgrade_steps.upgrade_from_800_801 import UpgradeFrom800To801
from swimlane_platform.upgrade_steps.upgrade_from_801_900 import UpgradeFrom801To900
from swimlane_platform.upgrade_steps.upgrade_from_900_910 import UpgradeFrom900To910
from swimlane_platform.upgrade_steps.upgrade_from_910_911 import UpgradeFrom910To911
from swimlane_platform.upgrade_steps.upgrade_from_911_912 import UpgradeFrom911To912
from swimlane_platform.upgrade_steps.upgrade_from_912_1000 import UpgradeFrom912To1000
from swimlane_platform.upgrade_steps.upgrade_from_1000_1001 import UpgradeFrom1000To1001
from swimlane_platform.upgrade_steps.upgrade_from_1001_1002 import UpgradeFrom1001To1002
from swimlane_platform.upgrade_steps.upgrade_from_1002_1010 import UpgradeFrom1002To1010
from swimlane_platform.upgrade_steps.upgrade_from_1010_1011 import UpgradeFrom1010To1011
from swimlane_platform.upgrade_steps.upgrade_from_1011_1012 import UpgradeFrom1011To1012
from swimlane_platform.upgrade_steps.upgrade_from_1012_1013 import UpgradeFrom1012To1013
from swimlane_platform.upgrade_steps.upgrade_from_1013_1020 import UpgradeFrom1013To1020
from swimlane_platform.upgrade_steps.upgrade_from_1020_1022 import UpgradeFrom1020To1022
from swimlane_platform.upgrade_steps.upgrade_from_1022_1030 import UpgradeFrom1022To1030
from swimlane_platform.upgrade_steps.upgrade_from_1030_1040rc2 import UpgradeFrom1030To1040rc2
from swimlane_platform.upgrade_steps.upgrade_from_1040rc2_1040rc3 import UpgradeFrom1040rc2To1040rc3
from swimlane_platform.upgrade_steps.upgrade_from_1040rc3_1040 import UpgradeFrom1040rc3To1040

Upgrades = [
    UpgradeFrom700To701,
    UpgradeFrom701To800,
    UpgradeFrom800To801,
    UpgradeFrom801To900,
    UpgradeFrom900To910,
    UpgradeFrom910To911,
    UpgradeFrom911To912,
    UpgradeFrom912To1000,
    UpgradeFrom1000To1001,
    UpgradeFrom1001To1002,
    UpgradeFrom1002To1010,
    UpgradeFrom1010To1011,
    UpgradeFrom1011To1012,
    UpgradeFrom1012To1013,
    UpgradeFrom1013To1020,
    UpgradeFrom1020To1022,
    UpgradeFrom1022To1030,
    UpgradeFrom1030To1040rc2,
    UpgradeFrom1040rc2To1040rc3,
    UpgradeFrom1040rc3To1040
]  # type: List[Type[UpgradeStep]]



from swimlane_platform import install
from swimlane_platform.lib import Configuration, Actions, BaseWithLog, ExpectedException, \
    automation_questions, logging_questions, dev_question
from swimlane_platform import upgrade
from swimlane_platform.backup import backup
from swimlane_platform.backup import restore


class SwimlaneWizard(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> None
        super(SwimlaneWizard, self).__init__(config)

    # noinspection PyBroadException
    def run(self):
        command = self.config.args.command
        try:
            if command == Actions.Install:
                install.run(self.config)
            elif command == Actions.Upgrade:
                upgrade.run(self.config)
            elif command == Actions.Backup:
                backup.run(self.config)
            elif command == Actions.Restore:
                restore.run(self.config)
        except ExpectedException as e:
            self.logger.error(e.message)
        except Exception:
            self.logger.exception("Unexpected error.")


def run():
    auto_config = Configuration()
    auto_config.collect(automation_questions)
    config = Configuration(auto_config.args.automation, auto_config.args.automation_file)
    questions = [
        {
            'type': 'list',
            'name': 'command',
            'message': 'What action do you want to perform?',
            'choices': [
                {
                    'name': 'Install:   Install Swimlane.',
                    'value': Actions.Install
                },
                {
                    'name': 'Upgrade:   Upgrade to a newer version of Swimlane.',
                    'value': Actions.Upgrade
                },
                {
                    'name': 'Backup:    Create a backup of Swimlane data.',
                    'value': Actions.Backup
                },
                {
                    'name': 'Restore:   Restore Swimlane data from backup.',
                    'value': Actions.Restore
                }
            ]
        }
    ]
    questions.extend(dev_question)
    questions.extend(logging_questions)
    config.collect(questions)
    SwimlaneWizard(config).run()



from os import path

from swimlane_platform.lib import BaseWithLog, DockerComposeFileManager, names, DockerImage
from abc import abstractproperty, ABCMeta, abstractmethod
import semver
from future.utils import with_metaclass


class UpgradeStep(with_metaclass(ABCMeta, BaseWithLog)):

    @abstractproperty
    def FROM(self):
        # type: () -> semver.VersionInfo
        pass

    @abstractproperty
    def TO(self):
        # type: () -> semver.VersionInfo
        pass

    @abstractmethod
    def process(self):
        # type: () -> None
        pass

    def upgrade_standard_images(self, dev, install_dir):
        # type: (bool, str) -> DockerComposeFileManager
        """

        :param dev: True if run in development environment.
        :param install_dir: Installation directory where docker-compose file is.
        :return: DockerComposeFileManager to add additional upgrades.
        """
        docker_compose = DockerComposeFileManager(self.logger, path.join(install_dir, names.DOCKER_COMPOSE_FILE))
        api_image = DockerImage(image_name='swimlane/swimlane-api', version=str(self.TO), is_dev=dev).full_name
        docker_compose.set(api_image, 'services', names.SW_API, 'image')
        tasks_image = DockerImage(image_name='swimlane/swimlane-tasks', version=str(self.TO), is_dev=dev).full_name
        docker_compose.set(tasks_image, 'services', names.SW_TASKS, 'image')
        web_image = DockerImage(image_name='swimlane/swimlane-web', version=str(self.TO), is_dev=dev).full_name
        docker_compose.set(web_image, 'services', names.SW_WEB, 'image')
        reports_image = DockerImage(image_name='swimlane/swimlane-reports', version=str(self.TO), is_dev=dev).full_name
        docker_compose.set(reports_image, 'services', names.SW_REPORTS, 'image')
        return docker_compose



from swimlane_platform.lib import Configuration, names, \
    BaseWithLog, info_function_start_finish, create_cert, ssl_questions
from os import path
from OpenSSL import crypto
from shutil import copy2


class EnableWebSsl(BaseWithLog):
    """
    Creates or enables existing certificates on Swimlane web.
    """

    def __init__(self, config):
        # type: (Configuration) -> EnableWebSsl
        super(EnableWebSsl, self).__init__(config)
        self.install_dir = names.INSTALL_DIR
        self.env_root = path.join(self.install_dir, names.SECRETS_SUB_FOLDER)
        self.secrets = path.join(self.install_dir, names.SECRETS_SUB_FOLDER)

    @info_function_start_finish('Enable SSL for Web')
    def run(self):
        # type: () -> None
        """
        Runs modifications to environment to enable additional settings for SSL
        """
        if self.config.args.web_ssl_self_signed:
            self.generate_certificates()
        else:
            # verify about names
            copy2(self.config.args.web_ssl_certificate, path.join(self.secrets, names.SSL_WEB_CERTIFICATE))
            copy2(self.config.args.web_ssl_key, path.join(self.secrets, names.SSL_WEB_KEY))

    def generate_certificates(self):
        # type: () -> None
        """
        Generates and saves web certificates
        """
        self.config.collect(ssl_questions)
        c, k = create_cert(country=self.config.args.ssl_country,
                           state=self.config.args.ssl_state,
                           location=self.config.args.ssl_location,
                           company=self.config.args.ssl_company,
                           application=self.config.args.ssl_application,
                           dns=self.config.args.ssl_dns)
        with open(path.join(self.secrets, names.SSL_WEB_KEY), "wb") as fw:
            fw.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k))
        with open(path.join(self.secrets, names.SSL_WEB_CERTIFICATE), "wb") as fw:
            fw.write(crypto.dump_certificate(crypto.FILETYPE_PEM, c))


def run(config):
    # type: (Configuration) -> None
    """
    The script run method, that can be called by other script or from file run method.
    :param config: Configuration information collected by parent script.
    """
    EnableWebSsl(config).run()



import shutil
from os import path
from swimlane_platform.backup.backup_restore_base import BackupRestoreBase
from swimlane_platform.lib import PathExistsValidator, \
    debug_function_args, names, info_function_start_finish


class Backup(BackupRestoreBase):

    @info_function_start_finish('Swimlane backup.')
    def run(self):
        # type: () -> None
        """
        Main method.
        """
        local, remote = self.get_archive_name(self.config.args.backup_dir)
        # noinspection PyBroadException
        try:
            self.docker_manager.containers_run(self.docker_manager.container_stop, names.SW_WEB, names.SW_TASKS,
                                               names.SW_API)
            self.backup_mongo(remote, self.DB_SWIMLANE)
            self.backup_mongo(remote, self.DB_SWIMLANE_HISTORY, True)
            self.report_archive_name(remote)
        except Exception:
            shutil.rmtree(local)
            raise
        finally:
            self.docker_manager.containers_run(self.docker_manager.container_start, names.SW_API, names.SW_TASKS,
                                               names.SW_WEB)

    @debug_function_args
    def backup_mongo(self, archive, database, auto_remove=False):
        # type: (str, str, bool) -> None
        """
        Backup MongoDb database.
        :param auto_remove: Remove container after image run.
        :param archive: This specific run folder. Archive.
        :param database: Database name.
        """
        self.logger.info("Started MongoDb {db} backup.".format(db=database))
        mongo_connection = self.get_mongo_connection()
        folder, cert_name = path.split(mongo_connection.pem_file)
        command = 'mongodump' \
                  ' --host {host}' \
                  ' --gzip' \
                  ' --excludeCollectionsWithPrefix hangfire' \
                  ' --db {db}' \
                  ' --archive={archive}/{db}.gz' \
                  ' --username {mongo_user}' \
                  ' --password {mongo_password}' \
                  ' --ssl' \
                  ' --sslPEMKeyFile {cert_folder}/{cert_name}' \
                  ' --sslAllowInvalidCertificates'.format(host=mongo_connection.host,
                                                          db=database,
                                                          mongo_user=mongo_connection.user_name,
                                                          mongo_password=mongo_connection.password,
                                                          archive=archive,
                                                          cert_name=cert_name,
                                                          cert_folder=self.CERT_FOLDER)
        volumes = {
            self.config.args.backup_dir: {'bind': self.BACKUP_FOLDER, 'mode': 'Z'},
            folder: {'bind': self.CERT_FOLDER, 'mode': 'rw'}
        }
        network = self.docker_manager.container_get_network_name(names.SW_MONGO)
        logs = self.client.containers.run(self.BACKUP_IMAGE,
                                          command=command,
                                          stderr=True,
                                          stream=True,
                                          auto_remove=auto_remove,
                                          volumes=volumes,
                                          network=network)
        for log in logs:
            self.logger.verbose(log)
        self.logger.info("Finished MongoDb {db} backup.".format(db=database))


def run(config):
    questions = [
        {
            'type': 'input',
            'name': 'backup_dir',
            'message': 'Folder for storing backups',
            'validate': PathExistsValidator
        }
    ]
    config.collect(questions)
    Backup(config).run()



from os import path
import subprocess
from semver import VersionInfo
from typing import List
from swimlane_platform.lib.version_manager import semver_parse
from swimlane_platform.lib.debug_decorators import debug_function_args, debug_function_return
from swimlane_platform.lib.logger import SplitStreamLogger
import re


class DockerComposeManager:

    def __init__(self, logger, *docker_compose_file):
        # type: (SplitStreamLogger, List[str]) -> None
        self.logger = logger
        self.docker_compose_files = docker_compose_file

    @debug_function_return
    @debug_function_args
    def docker_compose_run(self, command, *args):
        # type: (str, str) -> str
        """
        Runs docker compose commands.
        :param command: Docker compose command. (up, down, build etc)
        :param args: Command arguments
        """
        command_args = ['docker-compose']
        dir_name = ''
        for docker_compose_file in self.docker_compose_files:
            dir_name, file_name = path.split(docker_compose_file)
            command_args.extend(['-f', file_name])
        command_args.append(command)
        if args:
            command_args.extend(args)
        std_out = subprocess.check_output(command_args, cwd=dir_name)
        try:
            std_out = std_out.decode()
        except AttributeError:
            pass
        self.logger.verbose(std_out)
        return std_out

    @debug_function_return
    def docker_compose_down(self):
        # type: () -> str
        """
        Spins down services from docker compose.
        """
        return self.docker_compose_run('down')

    @debug_function_return
    def docker_compose_up(self, *services):
        # type: (List[str]) -> str
        """
        Starts services from docker compose.
        """
        return self.docker_compose_run('up', '-d', *services)

    @debug_function_return
    def docker_compose_version(self):
        # type: () -> VersionInfo
        """
        Gets docker-compose version
        :return: semver version of docker-compose
        """
        command_args = ['docker-compose', '--version']
        std_out = subprocess.check_output(command_args)
        try:
            std_out = std_out.decode()
        except AttributeError:
            pass
        self.logger.verbose(std_out)
        match = re.search(r'\b(?:\d+\.)*\d+\b', std_out)
        version = match.group() if match else '0'
        return semver_parse(version)



from typing import Tuple, Iterable, Dict, List, Union

__DELIMITER = '='


def __encode_line(key, value):
    # type: (str, str) -> str
    """
    Formats .env file line from key, value.
    :param key: Key.
    :param value: Value.
    :return: Line for .env file.
    """
    return '{key}{delimiter}{value}'.format(key=key, value=value, delimiter=__DELIMITER)


def __encode_dict(env_dict):
    # type: (Dict[str,str]) -> Iterable[str]
    """
    Generator of .env encoded strings.
    :param env_dict: Dictionary with environment values.
    """
    for key, value in env_dict.items():
        yield __encode_line(key, value)


def _parse_line(line):
    # type: (str) -> Tuple[str, Union[str, bool]]
    """
    Makes a two value tuple from = delimited environment variable
    :param line: environment variable
    :return: key, value - tuple
    """
    spl = str(line).rstrip('\n\r ').split(__DELIMITER)
    # is boolean
    if len(spl) == 2:
        value = spl[1]
        if value == 'False' or value == 'True':
            value = value == 'True'
        return str(spl[0]), value
    return str(spl[0]), str('='.join(spl[1:]))


def parse_env_variables(variables):
    # type: (List[str]) -> Iterable[Tuple[str, str]]
    """
    Parses the environment variable list from Docker inspect.
    :param variables: Formatted variable list.
    """
    for variable_line in variables:
        yield _parse_line(variable_line)


def parse_env_variables_to_dict(variables):
    # type: (List[str]) -> Dict[str, str]
    """
    Parses the environment variable list from Docker inspect.
    :param variables: Formatted variable list.
    """
    return dict(parse_env_variables(variables))


def read(filename):
    # type: (str) -> Iterable[Tuple[str, str]]
    """
    Generator of key, value pairs from file.
    :param filename: File name.
    """
    with open(filename, 'r') as fs:
        for line in fs.readlines():
            yield _parse_line(line)


def read_dict(filename):
    # type: (str) -> Dict[str, str]
    """
    Reads .env file as dictionary.
    :param filename: File name.
    :return: Dictionary of environment variables.
    """
    return dict(read(filename))


def write_dict(filename, env_dict):
    # type: (str, Dict[str, str]) -> None
    """
    Writes dictionary as .env file.
    :param filename: File name.
    :param env_dict: Dictionary with environment values.
    """
    with open(filename, 'w') as fs:
        encoded_dict = [line + '\n' for line in __encode_dict(env_dict)]
        encoded_dict.sort()
        fs.writelines(encoded_dict)





import shutil
from os import path
import copy
from typing import Callable, Any, Dict


def apply_for_all_dict_values(dictionary, valid, func):
    # type: (Dict, Callable[[Any], bool], Callable[[Any], Any]) -> None
    """Applies func to all values in Dictionary that match valid function.
    :param dictionary: The dictionary to iterate
    :param valid: Function that validates type or other condition for value.
    :param func: Function to apply to the value and replace it.
    """

    def _iterate(value):
        """Inner function with wider object definition then parent.
        Created to keep parent strict signature."""
        if isinstance(value, dict):
            for _, _value in value.items():
                _iterate(_value)
        elif isinstance(value, list):
            for _value in value:
                _iterate(_value)
        if valid(value):
            func(value)

    _iterate(dictionary)


def merge_dictionaries(first_dictionary, second_dictionary, overwrite=False):
    # type: (dict, dict, bool) -> dict
    """
    Merges content of two dictionaries including merging of lists and returns results
    :param first_dictionary: First dictionary
    :param second_dictionary: Second dictionary
    :param overwrite: If values are primitive whether to overwrite first dictionary value.
    :return: Merged dictionary result. If values were lists, lists are merged.
    """
    def _merge_dictionaries(d1, d2, replace):
        for key, value in d2.items():
            if key not in d1:
                d1[key] = value
            elif isinstance(d1[key], list) and isinstance(value, list):
                d1[key] += value
            elif isinstance(d1[key], dict):
                _merge_dictionaries(d1[key], value, replace)
            elif replace:
                d1[key] = value

    _first_dictionary = copy.deepcopy(first_dictionary)
    _merge_dictionaries(_first_dictionary, second_dictionary, overwrite)
    return _first_dictionary


def find_remote_sibling(location, name):
    folder = path.join(location, name)
    if path.exists(folder):
        return folder
    elif location == '/':
        return None
    else:
        return find_remote_sibling(path.dirname(location), name)


class FileSystemError(Exception):
    pass


def copy_if_not_exist(source, target):
    if path.exists(target):
        return
    if not path.exists(source):
        raise FileSystemError('{name} does not exist.'.format(name=source))
    shutil.copy2(source, target)



import functools

def funccode(func):
    try:
        return func.func_code
    except AttributeError:
        return func.__code__

def funcname(func):
    try:
        return func.func_name
    except AttributeError:
        return func.__name__

def info_function_start_finish(action_description=None):

    def wrap(func):
        action = action_description if action_description else funcname(func).replace('_', ' ') + '.'

        @functools.wraps(func)
        def echo_func(*args, **kwargs):
            if args[0] and hasattr(args[0], 'logger'):
                logger = args[0].logger
                logger.info('Starting {action}'.format(action=action))
                result = func(*args, **kwargs)
                logger.info('Finishing {action}'.format(action=action))
                return result
            else:
                return func(*args, **kwargs)

        return echo_func
    return wrap


def verbose_function_start_finish(action_description=None):

    def wrap(func):
        action = action_description if action_description else funccode(func).replace('_', ' ') + '.'

        @functools.wraps(func)
        def echo_func(*args, **kwargs):
            if args[0] and hasattr(args[0], 'logger'):
                logger = args[0].logger
                try:
                    logger.verbose('Starting {action}'.format(action=action))
                    result = func(*args, **kwargs)
                    logger.verbose('Finishing {action}'.format(action=action))
                    return result
                except AttributeError:
                    pass
            return func(*args, **kwargs)

        return echo_func
    return wrap


def debug_function_args(func):
    arg_names = funccode(func).co_varnames[:funccode(func).co_argcount]
    function_name = funcname(func)

    @functools.wraps(func)
    def echo_func(*args, **kwargs):
        if args[0] and hasattr(args[0], 'logger'):
            logger = args[0].logger
            names = arg_names[1:]
            values = args[1:]
            asterisk = []
            if len(names) < len(values):
                asterisk = [('*', values[len(names):])]
            args_with_values = ', '.join('%s=%r' % entry for entry in list(zip(names, values)) + asterisk + list(kwargs.items()))
            logger.debug('{name} called with {args}'.format(name=function_name, args=args_with_values))

        return func(*args, **kwargs)

    return echo_func


def debug_function_return(func):
    function_name = funcname(func)

    @functools.wraps(func)
    def echo_func(*args, **kwargs):
        if args[0]and hasattr(args[0], 'logger'):
            logger = args[0].logger
            result = func(*args, **kwargs)
            logger.debug('{name} returned {result}'.format(name=function_name, result=result))
            return result
        else:
            return func(*args, **kwargs)

    return echo_func



from argparse import Namespace, ArgumentParser, SUPPRESS
from typing import Dict, List, Any
from PyInquirer import prompt, Validator, ValidationError
from swimlane_platform.lib.models import Automation
from swimlane_platform.lib.version_manager import semver_parse
from os import path
from validators.domain import domain
from validators.ip_address import ipv4, ipv6
import json


class AnswerRequiredValidator(Validator):
    def validate(self, document):
        if not document or not document.text:
            raise ValidationError(
                message='Entry required.',
                cursor_position=len(document.text))  # Move cursor to end


class VersionValidator(AnswerRequiredValidator):
    def validate(self, document):
        super(VersionValidator, self).validate(document)
        if document.text in ['dev', 'develop', 'latest']:
            return
        version = semver_parse(document.text)
        if not version:
            raise ValidationError(message='Version provided is invalid semver and cannot be parsed.',
                                  cursor_position=len(document.text))


class DnsValidator(AnswerRequiredValidator):
    def validate(self, document):
        super(DnsValidator, self).validate(document)
        if ' ' in document.text:
            raise ValidationError(message='DNS should be comma-separated and cannot contain spaces.',
                                  cursor_position=len(document.text))
        for dns in document.text.split(","):
            if domain(dns) or ipv4(dns) or ipv6(dns):
                continue
            elif dns[0] == "*" and domain(dns.lstrip("*.")):
                continue
            raise ValidationError(message='DNS should be valid urls or ip addresses.',
                                  cursor_position=len(document.text))


class PathExistsValidator(AnswerRequiredValidator):
    def validate(self, document):
        super(PathExistsValidator, self).validate(document)
        if not path.exists(document.text):
            raise ValidationError(
                message='Path doesn\'t exist.',
                cursor_position=len(document.text))  # Move cursor to end


class ContainingDirectoryExistsValidator(AnswerRequiredValidator):
    def validate(self, document):
        super(ContainingDirectoryExistsValidator, self).validate(document)
        if not path.exists(path.dirname(document.text)):
            raise ValidationError(
                message='Path doesn\'t exist.',
                cursor_position=len(document.text))  # Move cursor to end


class LogFileOptional(Validator):
    def validate(self, document):
        if not document or not document.text:
            return
        if not path.exists(path.dirname(document.text)):
            raise ValidationError(
                message='Directory doesn\'t exist.',
                cursor_position=len(document.text))  # Move cursor to end


class Arguments(Namespace):

    def __init__(self, **kwargs):
        super(Arguments, self).__init__(**kwargs)

    def __getattr__(self, item):
        return None


class Configuration(dict):

    def __init__(self, action=Automation.Normal, file_path=None):
        # type: (str, str) -> None
        super(Configuration, self).__init__()
        self.action = action
        self.file_path = file_path

    def add_other_source(self, other_config, overwrite=False):
        # type: (Dict[str, Any], bool) -> None
        """
        Ability to add other configurations, from environment variables for instance.
        :param other_config: Other configurations. Dictionary.
        :param overwrite: Whether to add it to existing or overwrite them.
        """
        for key, value in other_config.items():
            if key not in self or overwrite:
                self[key] = value

    def collect(self, questions):
        # type: (List[Dict[str, str]]) -> None
        self._add_question_lookup_in_args(questions)
        self._load_config_file()
        remaining_questions = [q for q in questions if not q.get('name') in self]
        self.update(prompt(remaining_questions, answers=self))
        self._save_config_file()

    def _save_config_file(self):
        # type: () -> None
        """
        If user requested to record session, saves answers to the file.
        """
        if self.action == Automation.Save:
            assert self.file_path
            if not path.exists(self.file_path):
                self._write_dict(self.file_path, self)
            else:
                existing = self._read_dict(self.file_path)
                for key, value in self.items():
                    existing[key] = value
                self._write_dict(self.file_path, existing)

    def _load_config_file(self):
        # type: () -> None
        """
        If users request to load the file, then it reads it and updates inner dictionary.
        """
        if self.action == Automation.Load:
            assert self.file_path
            config = self._read_dict(self.file_path)
            self.update(config)

    @property
    def args(self):
        # type: () -> Arguments
        """
        Answers in the type of Arguments. Values can be found by key or attribute.
        Returns new object each time. Do not assign it to a variable and expect it being updated.
        :return: Answers as Arguments.
        """
        return Arguments(**self)

    def _add_question_lookup_in_args(self, questions):
        # type: (List[Dict[str, str]]) -> None
        """
        Checks if the questions has been answered in cli arguments.
        :param questions: Questions to check.
        """
        parser = ArgumentParser()

        def get_argument_type(question_type):
            if question_type == 'input' or question_type == 'password' or question_type == 'list':
                return 'store'
            elif question_type == 'confirm':
                return 'store_true'
            else:
                raise Exception("The question type is not mapped.")

        for question in questions:
            parser.add_argument('--{name}'.format(name=question.get('name')),
                                help=question.get('message'),
                                default=SUPPRESS,
                                action=get_argument_type(question.get('type')))
        args, _ = parser.parse_known_args()
        self.update(dict(args.__dict__))

    @staticmethod
    def _write_dict(file_path, _dict):
        # type: (str, Dict[str, Any]) -> None
        """
        Writes dictionary as a json file.
        :param file_path: File path to save json to.
        :param _dict: Dictionary to save.
        """
        assert file_path
        assert _dict
        with open(file_path, 'w') as fp:
            json.dump(_dict, fp, indent=2, separators=(',', ': '))

    @staticmethod
    def _read_dict(file_path):
        # type: (str) -> Dict[str: Any]
        """
        Reads dictionary from json file
        :param file_path: File path to read json from.
        """
        assert file_path
        with open(file_path, 'r') as fp:
            return json.load(fp)



from __future__ import print_function
from swimlane_platform.lib.os_manager import OsManager
from swimlane_platform.shared_steps import set_mongo_tls_flags
from swimlane_platform.lib import Configuration, names, DockerComposeFileManager, \
    read_dict, write_dict, BaseWithLog, info_function_start_finish, debug_function_args, \
    debug_function_return, create_cert, logging_questions, ssl_questions
from os import path, makedirs
import re
from OpenSSL import crypto
from shutil import copy2
from typing import Dict
from swimlane_platform.lib.db_configuration import DbConfiguration


class EnableMongoSsl(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> EnableMongoSsl
        super(EnableMongoSsl, self).__init__(config)
        self.install_dir = names.INSTALL_DIR
        self.env_root = path.join(self.install_dir, names.SECRETS_SUB_FOLDER)
        self.secrets = path.join(self.install_dir, names.SECRETS_SUB_FOLDER)
        self.db_configuration = DbConfiguration(self.config)
        self.os_manager = OsManager(self.logger)

    @info_function_start_finish('Enable SSL for MongoDB')
    def run(self):
        # type: () -> None
        """
        Runs modifications to environment to enable additional settings for SSL
        """
        if not path.exists(self.secrets):
            makedirs(self.secrets)
        if self.config.args.mongo_ssl_self_signed:
            self._env_modify(False)
            cert = self._generate_certificates()
            self.os_manager.copy_with_permissions(cert, cert, 0o644, True)
            self._set_secret_source_docker_compose(cert)
        else:
            self._env_modify(True)
            url = self._get_certificate_url(self.config.args.certificate)
            _, file_name = path.split(self.config.args.certificate)
            file_destination = path.join(self.secrets, file_name)
            self.os_manager.copy_with_permissions(self.config.args.certificate, file_destination, 0o644, True)
            self._add_alias_to_docker_compose(url)
            self._set_secret_source_docker_compose(self.config.args.certificate)
            self._env_change_server_name(url)
        set_mongo_tls_flags.run(self.config)

    @debug_function_args
    def _env_modify(self, verify_certificate):
        # type: (bool) -> None
        self._env_modify_file(names.API_ENV_FILE, verify_certificate)
        self._env_modify_file(names.TASKS_ENV_FILE, verify_certificate)

    @debug_function_args
    def _env_modify_file(self, file_name, verify_certificate):
        # type: (str, bool) -> None
        api_env_file = path.abspath(path.join(self.env_root, file_name))
        api_env = read_dict(api_env_file)
        self._env_enable_mongo_ssl(api_env, 'SwimlaneConnectionString', verify_certificate)
        self._env_enable_mongo_ssl(api_env, 'HistoryConnectionString', verify_certificate)
        write_dict(api_env_file, api_env)

    @debug_function_args
    def _add_alias_to_docker_compose(self, certificate_url):
        # type: (str) -> None
        docker_compose = DockerComposeFileManager(self.logger, path.join(self.install_dir, 'docker-compose.yml'))
        alias_path = ['services', names.SW_MONGO, 'networks', 'internal_network', 'aliases']
        docker_compose.append_or_create_list(certificate_url, *alias_path)
        docker_compose.save()

    @debug_function_args
    def _set_secret_source_docker_compose(self, host_cert_path):
        # type: (str) -> None
        _, file_name = path.split(host_cert_path)
        install_cert_path = path.join(self.secrets, file_name)
        docker_compose = DockerComposeFileManager(self.logger, path.join(self.install_dir, 'docker-compose.yml'))
        alias_path = ['secrets', 'mongodb.pem',  'file']
        docker_compose.set(install_cert_path, *alias_path)
        docker_compose.append_or_create_list('mongodb.pem', 'services', names.SW_MONGO, 'secrets')
        docker_compose.save()

    @debug_function_args
    def _env_enable_mongo_ssl(self, api_env, conn_name_part, verify_certificate):
        # type: (Dict[str, str], str, bool) -> None
        regex = '.*{name}'.format(name=conn_name_part)
        for key, value in api_env.items():
            if re.match(regex, key):
                mongo_settings = self.db_configuration.parse_mongo_uri(value)
                options = mongo_settings['options']
                options['ssl'] = str(True).lower()
                options['sslVerifyCertificate'] = str(verify_certificate).lower()
                api_env[key] = self.db_configuration.get_mongo_uri(mongo_settings)

    @debug_function_return
    @debug_function_args
    def _generate_certificates(self, one_file=True):
        # type: (bool) -> str
        file_template = path.join(self.secrets, 'mongodb.{ext}')
        self.config.collect(ssl_questions)
        cert, k = create_cert(country=self.config.args.ssl_country,
                              state=self.config.args.ssl_state,
                              location=self.config.args.ssl_location,
                              company=self.config.args.ssl_company,
                              application=self.config.args.ssl_application,
                              dns=self.config.args.ssl_dns)
        if one_file:
            pem_file = file_template.format(ext='pem')
            open(pem_file, "wb").write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k))
            open(pem_file, "ab").write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
            return pem_file
        else:
            open(file_template.format(ext='key'), "wb").write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k))
            open(file_template.format(ext='crt'), "wb").write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
            return file_template.format(ext='crt')

    @debug_function_return
    @debug_function_args
    def _get_certificate_url(self, certificate_location):
        # type: (str) -> str
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, open(certificate_location, 'rt').read())
        return cert.get_subject().CN

    @debug_function_args
    def _env_change_server_name(self, url):
        # type: (str) -> None
        for file_name in [names.API_ENV_FILE, names.TASKS_ENV_FILE]:
            env_file = path.abspath(path.join(self.env_root, file_name))
            env = read_dict(env_file)
            connection_names = ['SwimlaneConnectionString', 'HistoryConnectionString']
            for regex in ['.*{name}'.format(name=name) for name in connection_names]:
                for key, value in env.items():
                    if re.match(regex, key):
                        connection = self.db_configuration.parse_mongo_uri(value)
                        new_node_list = []
                        for _, port in connection['nodelist']:
                            new_node_list.append((url, port))
                        connection['nodelist'] = new_node_list
                        env[key] = self.db_configuration.get_mongo_uri(connection)
            write_dict(env_file, env)


def run(config):
    # type: (Configuration) -> None
    """
    The script run method, that can be called by other script or from file run method.
    :param config: Configuration information collected by parent script.
    """
    questions = [
        {
            'type': 'confirm',
            'message': 'Do you want to use a self-signed certificate for SSL connections to MongoDB? '
                       'If not, you must provide your own certificate in .pem format.',
            'name': 'mongo_ssl_self_signed',
            'default': True
        },
        {
            'type': 'input',
            'name': 'certificate',
            'message': 'Specify the full path to your .pem format certificate.',
            'when': lambda a: not a['mongo_ssl_self_signed']
        }
    ]
    questions.extend(logging_questions)
    config.collect(questions)
    EnableMongoSsl(config).run()



from os import path
import os
import pkg_resources
import swimlane_platform
from swimlane_platform.lib.functions import find_remote_sibling
from swimlane_platform.lib import ValidationException, \
    Configuration, \
    info_function_start_finish, \
    debug_function_args, \
    BaseWithLog, \
    names, \
    DockerComposeManager, \
    VersionValidator, \
    DockerManager, \
    DockerComposeFileManager, \
    web_ssl_questions, \
    ssl_questions
from swimlane_platform.shared_steps import enable_mongo_ssl
from swimlane_platform.lib.version_manager import pypi_version_to_semver
from swimlane_platform.lib.os_manager import OsManager
from swimlane_platform.lib.constants import SWIMLANE_UID_GID
from swimlane_platform.install_steps import enable_web_ssl
from swimlane_platform.add_file_encryption_key import run as run_add_file_encryption
from swimlane_platform.install_steps.setup_swimlane_environment import run as run_setup_environment
from swimlane_platform.install_steps.set_permissions import run as run_set_permissions
from swimlane_platform.shared_steps import enable_turbine
from swimlane_platform.install_steps.setup_dbs import run as run_setup_dbs
from swimlane_platform.lib.questions_groups import offline_questions
import shutil
import glob
from swimlane_platform.install_steps import PreFlightCheck
from builtins import str


class SwimlaneInstaller(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> None
        super(SwimlaneInstaller, self).__init__(config)
        script_path = path.dirname(path.realpath(__file__))
        self.template_dir = find_remote_sibling(script_path, names.TEMPLATE_DIR)
        self.docker_manager = DockerManager(self.logger)
        self.os_manager = OsManager(self.logger)

    @info_function_start_finish('Install Swimlane.')
    def run(self):
        # type: () -> None
        """
        Main Swimlane install method.
        """
        if not PreFlightCheck(self.config).run():
            return
        self.install_swimlane_files()
        self.get_swimlane_images()
        # order is important
        enable_web_ssl.run(self.config)
        enable_mongo_ssl.run(self.config)
        # ******************
        run_setup_environment(self.config)
        run_setup_dbs(self.config)
        run_set_permissions(self.config)
        run_add_file_encryption(self.config)
        enable_turbine.run(self.config)
        self.run_after_install()

    @info_function_start_finish('Installing Swimlane files.')
    def install_swimlane_files(self):
        """
        Make install directories and copy files there
        """
        self.create_if_does_not_exists(names.INSTALL_DIR)
        secrets_dir = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)
        self.create_if_does_not_exists(secrets_dir)
        db_init_dir = path.join(names.INSTALL_DIR, names.DB_INIT_SUB_FOLDER)
        self.create_if_does_not_exists(db_init_dir)
        self.copy_if_does_not_exists(self.template_dir, names.INSTALL_DIR, '*.yml')
        template_db_init = path.join(self.template_dir, names.DB_INIT_SUB_FOLDER)
        self.copy_if_does_not_exists(template_db_init, db_init_dir, '*.sh')
        self.copy_if_does_not_exists(path.join(self.template_dir, names.SECRETS_SUB_FOLDER), secrets_dir, '.*')
        # Explicit name due to security issues.
        self.os_manager.copy_with_permissions(path.join(template_db_init, 'init-mongodb-users.sh'),
                                              path.join(db_init_dir, 'init-mongodb-users.sh'), 0x655)

    @info_function_start_finish('Getting Swimlane docker images.')
    def get_swimlane_images(self):
        """Loads or pulls docker images required by Swimlane.
        In case of offline, analyses the images bundled, until build will know or care about repository
        re-tags nexus images. In case of online, analyses docker-compose file template and decides
        which version to pull.
        In both cases modifies docker-compose template to use correct images.
        """
        docker_compose_file = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE)
        docker_compose_file_manager = DockerComposeFileManager(self.logger, docker_compose_file)
        version = str(self.config.args.version) if self.config.args.version else self.latest_version()
        repository, version = self.docker_manager.extract_or_pull_images(docker_compose_file_manager.get_image_tags(),
                                                                         version,
                                                                         self.config.args.dev,
                                                                         self.config.args.installer_is_offline,
                                                                         self.config.args.extracted_files_folder)
        docker_compose_file_manager.replace_in_file_and_reload(names.TEMPLATE_REPOSITORY_TAG, repository)
        docker_compose_file_manager.replace_in_file_and_reload(names.TEMPLATE_VERSION_TAG, version)

    @staticmethod
    def latest_version():
        # type: () -> str
        """
        Returns the latest version of swimlane platform.
        :return: swimlane platform package version.
        """
        version_string = pkg_resources.get_distribution(swimlane_platform.__name__).version
        version = pypi_version_to_semver(version_string)
        if version:
            return str(version)
        else:
            raise ValidationException('Cannot parse package version {version} to Swimlane'.format(version=version))

    @info_function_start_finish('After installation.')
    def run_after_install(self):
        # type: () -> None
        """
        Prompts the user and starts or leaves containers as is.
        """
        START_COMMAND_MSG = 'cd /opt/swimlane and run "docker-compose up -d"'
        if not self.os_manager.swimlane_can_start():
            self.logger.warn('Swimlane is not currently able to start. Please change the UID and GID of the ' +
                             'current user to {uid_gid}, then {start_cmd} to start the Swimlane Services manually'
                             .format(uid_gid=SWIMLANE_UID_GID, start_cmd=START_COMMAND_MSG))
            return

        questions = [{
            'type': 'confirm',
            'name': 'start_swimlane',
            'message': 'Would you like to start the Swimlane services?'
        }]

        self.config.collect(questions)
        if self.config.args.start_swimlane:
            docker_compose_file = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE)
            DockerComposeManager(self.logger, docker_compose_file).docker_compose_up()
        else:
            self.logger.info(
                'To start the Swimlane Services manually, {0}.'.format(START_COMMAND_MSG))

    @debug_function_args
    def create_if_does_not_exists(self, dir_path):
        """Creates directory first checking for it's existence."""
        if not path.exists(dir_path):
            os.mkdir(dir_path)
            self.logger.verbose("{source} directory created.".format(source=dir_path))

    @debug_function_args
    def copy_if_does_not_exists(self, source, target, mask):
        for filepath in glob.glob1(source, mask):
            if not path.exists(path.join(target, filepath)):
                source_file = path.join(source, filepath)
                shutil.copy2(source_file, target)
                self.logger.verbose("File {source} copied to {target}".format(source=source_file, target=target))


def run(config):
    # type: (Configuration) -> None
    # noinspection PyCompatibility
    questions = offline_questions
    questions.append({
        'type': 'input',
        'name': 'version',
        'message': 'For development installation which version do you want to use?',
        'default': SwimlaneInstaller.latest_version(),
        'validate': VersionValidator,
        'when': lambda a: a.get('dev') and not a.get('installer_is_offline')
    })
    questions.extend(web_ssl_questions)
    config.collect(questions)
    if config.args.web_ssl_self_signed:
        config.collect(ssl_questions)
    SwimlaneInstaller(config).run()



#!/usr/bin/env python
from os import path
from semver import VersionInfo
from typing import List
from swimlane_platform.lib import Configuration, DockerComposeFileManager, DockerComposeManager, DockerManager, \
    names, debug_function_args, debug_function_return, info_function_start_finish, \
    ValidationException, BaseWithLog
from swimlane_platform.lib.args_config_questions import PathExistsValidator, VersionValidator
from swimlane_platform.lib.version_manager import semver_parse
from swimlane_platform.upgrade_steps import Upgrades
from swimlane_platform.lib.questions_groups import offline_questions
import semver


class SwimlaneUpgrader(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> None
        super(SwimlaneUpgrader, self).__init__(config)
        self.docker_manager = DockerManager(self.logger)

    @info_function_start_finish('Swimlane upgrade.')
    def run(self):
        # type: () -> None
        """
        Main method for upgrading Swimlane installation.
        """
        args = self.config.args
        current_version = self.get_service_version(names.SW_API)
        requested_version = semver.parse_version_info(args.version)
        if not self.need_upgrade(current_version, requested_version):
            self.logger.info('No upgrade applied, current version is bigger or equal to requested one.')
            return

        compose_file = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE)
        compose_file_override = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_OVERRIDE_FILE)
        old_images = DockerComposeFileManager(self.logger, compose_file).get_image_tags()
        self.stop_containers(compose_file, compose_file_override)
        self.run_incremental_upgrades(current_version, requested_version)
        new_images = DockerComposeFileManager(self.logger, compose_file).get_image_tags()
        self.docker_manager.extract_or_pull_images(new_images,
                                                   args.version,
                                                   args.dev,
                                                   args.installer_is_offline,
                                                   args.extracted_files_folder)
        self.clean_images(old_images, new_images)
        self.start_containers(compose_file, compose_file_override)

    @debug_function_return
    @debug_function_args
    def need_upgrade(self, current_version, upgrade_version):
        # type: (semver.VersionInfo, semver.VersionInfo) -> bool
        """
        Compares versions to determine if upgrade is needed.
        :param current_version: Current version of Swimlane.
        :param upgrade_version: New version of Swimlane.
        :return: True if upgrade needed.
        """
        return current_version < upgrade_version

    @debug_function_return
    @debug_function_args
    def get_service_version(self, service_name):
        # type: (str) -> semver.VersionInfo
        """
        Gets service image version.
        :return: image version
        """
        docker_compose = DockerComposeFileManager(self.logger, path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE))
        image_name = docker_compose.get('services', service_name, 'image')
        return self._get_version_from_image(image_name)

    @info_function_start_finish()
    @debug_function_args
    def clean_images(self, old_images, new_images):
        # type: (List[str],List[str]) -> None
        """
        Removes old images if they are replaced with newer ones.
        """
        images_to_remove = (image for image in old_images if image not in new_images)
        for old_image in images_to_remove:
            self.docker_manager.image_remove(old_image)
            self.logger.info('Removed {image} image.'.format(image=old_image))

    @info_function_start_finish('Running incremental upgrades.')
    @debug_function_args
    def run_incremental_upgrades(self, current_version, version):
        # type: (VersionInfo, VersionInfo) -> None
        """
        Runs all incremental upgrades up to a particular version.
        :param current_version: Current version.
        :param version: Target version.
        """
        if not self.need_upgrade(current_version, version):
            return
        upgrade = next((u for u in Upgrades if u.FROM == current_version), None)
        if upgrade:
            upgrade_step = upgrade(self.config)
            upgrade_step.process()
            self.run_incremental_upgrades(upgrade_step.TO, version)

    @debug_function_return
    @debug_function_args
    def _get_version_from_image(self, image_name):
        # type: (str) -> semver.VersionInfo
        """
        Returns the version part of an image.
        :param image_name:
        :return:
        """
        error = ValidationException('Cannot get current version. Cannot determine steps needed.')
        if not image_name:
            raise error
        split = image_name.split(':')
        if len(split) < 1:
            raise error
        tag = split[-1]
        version = semver_parse(tag)
        if not version:
            if tag == '4.0-7.0.0-beta':
                return VersionInfo(7, 0, 0)
            elif tag == '4.0-7.0.1-beta':
                return VersionInfo(7, 0, 1)
            raise error
        return version

    @info_function_start_finish()
    @debug_function_args
    def start_containers(self, *args):
        # type: (List[str]) -> None
        """
        Starts containers
        :param args: docker-compose files (full path).
        """
        DockerComposeManager(self.logger, *args).docker_compose_up()

    @info_function_start_finish()
    @debug_function_args
    def stop_containers(self, *args):
        # type: (List[str]) -> None
        """
        Starts containers
        :param args: docker-compose files (full path).
        """
        DockerComposeManager(self.logger, *args).docker_compose_down()


def run(config):
    # type: (Configuration) -> None
    """
    The script run method, that can be called by other script.
    :param config: Configuration information collected by parent script.
    """
    questions = offline_questions
    questions.append(
        {
            'type': 'input',
            'name': 'version',
            'message': 'What version do you want to upgrade to?',
            'validate': VersionValidator
        })
    config.collect(questions)
    SwimlaneUpgrader(config).run()



from swimlane_platform.lib.names import names
from swimlane_platform.lib.debug_decorators import info_function_start_finish
from swimlane_platform.shared_steps import enable_turbine
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
import semver


class UpgradeFrom1012To1013(UpgradeStep):
    FROM = semver.parse_version_info('10.1.2')  # type: semver.VersionInfo
    TO = semver.parse_version_info('10.1.3')  # type: semver.VersionInfo

    @info_function_start_finish('Upgrade From 10.1.2 To 10.1.3.')
    def process(self):
        # type: () -> None
        enable_turbine.run(self.config)
        self.upgrade_standard_images(self.config.args.dev, names.INSTALL_DIR).save()



from semver import VersionInfo

from swimlane_platform.lib import info_function_start_finish
from swimlane_platform.upgrade_steps.upgrade_step import UpgradeStep
import semver


class UpgradeFrom700To701(UpgradeStep):
    FROM = semver.parse_version_info('7.0.0')  # type: VersionInfo
    TO = semver.parse_version_info('7.0.1')  # type: VersionInfo

    @info_function_start_finish('Upgrade From 7.0.0 To 7.0.1')
    def process(self):
        # type: () -> None
        pass



import platform
import distro
from typing import Tuple
from swimlane_platform.lib import names
from swimlane_platform.lib.version_manager import semver_parse
from swimlane_platform.lib.logger import SplitStreamLogger
from swimlane_platform.lib.debug_decorators import debug_function_args, debug_function_return
from swimlane_platform.lib.constants import SWIMLANE_UID_GID
from os import stat, chmod, path, listdir
import os

try:
    # noinspection PyUnresolvedReferences
    from os import chown
except ImportError:
    pass


class OsManager:

    def __init__(self, logger):
        # type: (SplitStreamLogger) -> OsManager
        self.logger = logger

    @debug_function_return
    @debug_function_args
    def is_linux(self):
        # type: () -> bool
        """
        Verifies that os is linux
        :return: True if linux
        """
        return platform.system() == 'Linux'

    @debug_function_return
    @debug_function_args
    def is_match_distribution_linux(self, name, min_version, major_version_must_match=False):
        # type: (str, str, bool) -> bool
        """
        Verifies that os matches parameters send.
        :param name: Distribution name. i.e centos, rhel etc
        :param min_version: Minimum version of distribution
        :param major_version_must_match: Whether distro's major version must match that of min_version 
        :return: True if matches
        """
        linux_distribution = platform.linux_distribution(full_distribution_name=False)
        if not linux_distribution:
            self.logger.debug('Cannot obtain linux distribution info.')
            return False
        distribution_name, distribution_version, _ = linux_distribution
        if name != distribution_name:
            self.logger.debug('{value} does not match the name.'.format(value=distribution_name))
            return False
        version_info = semver_parse(distribution_version)
        if semver_parse(min_version) > version_info:
            self.logger.debug('{d_version}:{p_version} is lower than min version.'
                              .format(d_version=distribution_version, p_version=str(version_info)))
            return False
        if major_version_must_match and semver_parse(min_version).major != version_info.major:
            self.logger.debug('{d_version}:{p_version} has different major version than min version.'
                                .format(d_version=distribution_version, p_version=str(version_info)))
            return False
        return True

    @debug_function_return
    @debug_function_args
    def get_platform(self):
        # type: () -> str
        """
        Gets platform os long name. i.e. Linux-3.10.0-957.1.3.el7.x86_64-x86_64-with-centos-7.6.1810-Core
        :return: Long name
        """
        return platform.platform()

    @debug_function_return
    @debug_function_args
    def get_permissions_linux(self, file_system_path):
        # type: (str) -> int
        """
        Gets linux permissions of file system objects as oct str. i.e 660 etc
        :param file_system_path: File or folder path.
        :return Octal number for permission
        """
        return stat(file_system_path).st_mode & 0o777

    @debug_function_args
    def set_permissions_linux(self, file_system_path, permission):
        # type: (str, oct) -> None
        """
        Sets permissions on linux on file system objects.
        :param file_system_path: File or folder path.
        :param permission: Octal number for permission.
        """
        chmod(file_system_path, permission)

    @debug_function_return
    @debug_function_args
    def get_owner_linux(self, file_system_path):
        # type: (str) -> Tuple[int, int]
        """
        Gets linux owner information of file system objects as int. i.e (501, 500) etc
        :param file_system_path: File or folder path.
        :return Tuple with uid and gid respectively
        """
        stats = stat(file_system_path)
        return stats.st_uid, stats.st_gid

    @debug_function_args
    def set_owner_linux(self, file_system_path, uid, gid):
        # type: (str, int, int) -> None
        """
        Sets owner on linux on file system objects.
        :param file_system_path: File or folder path.
        :param uid: User id.
        :param gid: Group id.
        """
        chown(file_system_path, uid, gid)

    @debug_function_return
    def get_distribution_linux(self):
        # type: () -> str
        """
        Returns Linux distribution id. i.e centos, rhel etc
        :return: distribution id.
        """
        return distro.id()

    @debug_function_return
    def swimlane_can_start(self):
        """
        Checks whether or not Swimlane can start under the current user
        :return True if Swimlane can start under the current user, False otherwise
        """
        secrets_dir = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)
        files_to_check = {names.DB_ENCRYPTION_KEY, names.SSL_WEB_CERTIFICATE, names.SSL_WEB_KEY}
        for filename in filter(lambda f: f in files_to_check, listdir(secrets_dir)):
            stats = stat(path.join(secrets_dir, filename))
            if stats.st_uid != SWIMLANE_UID_GID:
                return False
        return True

    @debug_function_args
    def copy_with_permissions(self, file_origination, file_destination, permission, is_binary=False):
        """
        Copies content of the file to the file with different permissions.
        @param file_origination: Full path to original file.
        @param file_destination: Full path to destination file.
        @param permission: Permissions in numeric format.
        """
        # Getting content early, so you can overwrite itself.
        with open(file_origination, ('rb' if is_binary else 'r')) as file_out:
            content = file_out.read()

        umask = 0o777 ^ permission
        try:
            os.remove(file_destination)
        except OSError:
            pass

        # Open file descriptor
        umask_original = os.umask(umask)
        try:
            file_destination = os.open(file_destination, os.O_WRONLY | os.O_CREAT | os.O_EXCL, permission)
        finally:
            os.umask(umask_original)

        # Open file handle and write to file
        with os.fdopen(file_destination, ('wb' if is_binary else 'w')) as file_in:
            file_in.write(content)



from setuptools import setup, Command
from distutils.command.register import register as register_orig
from distutils.command.upload import upload as upload_orig
import shutil
import os
import sys
from glob import glob
from json import load


def convert(value):
    """Converts some specific json objects to python object"""
    if isinstance(value, dict):
        return {convert(k): convert(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [convert(element) for element in value]
    elif sys.version_info[0] == 2 and isinstance(value, unicode):
        return value.encode("utf-8")
    else:
        return value


def readme():
    with open('README.md') as f:
        return f.read()


def requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()


def super_glob(*path):
    files = []
    for arg in path:
        files.extend(glob(arg))
    return files


class Register(register_orig):

    def _get_rc_file(self):
        return os.path.join('.', '.pypirc')


class Upload(upload_orig):

    def _get_rc_file(self):
        return os.path.join('.', '.pypirc')


class Clean(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    # noinspection PyMethodMayBeStatic
    def run(self):
        shutil.rmtree('build', ignore_errors=True)
        shutil.rmtree('swimlane_platform.egg-info', ignore_errors=True)


with open('setup.json', 'r') as config_file:
    config = load(config_file, object_hook=convert)

setup(
    packages=['swimlane_platform',
              'swimlane_platform.lib',
              'swimlane_platform.install_steps',
              'swimlane_platform.shared_steps',
              'swimlane_platform.upgrade_steps',
              'swimlane_platform.backup'],
    long_description=readme(),
    long_description_content_type="text/markdown",
    install_requires=requirements(),
    include_package_data=True,
    cmdclass={
        'clean': Clean,
        'register': Register,
        'upload': Upload
    },
    entry_points={
        'console_scripts': ['swimlane-platform=swimlane_platform.wizard:run'],
    },
    data_files=[
        ('', ['setup.json']),
        ('swimlane_template_dir', glob('swimlane_template_dir/*.yml')),
        ('swimlane_template_dir/db-init', glob('swimlane_template_dir/db-init/*.sh')),
        ('swimlane_template_dir/.secrets', glob('swimlane_template_dir/.secrets/.*'))
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6"
    ],
    **config
)



from __future__ import print_function
from swimlane_platform.lib import names
from swimlane_platform.lib.base_with_log import BaseWithLog
from swimlane_platform.lib.docker_manager import DockerManager
from time import time
from swimlane_platform.lib.docker_compose_manager import DockerComposeManager
from swimlane_platform.lib.expected_exceptions import ExpectedException
from swimlane_platform.lib.args_config_questions import Configuration
from os import path
from swimlane_platform.lib.debug_decorators import info_function_start_finish, verbose_function_start_finish


class SetupDbs(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> SetupDbs
        super(SetupDbs, self).__init__(config)
        self.secrets_dir = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)
        self.docker_manager = DockerManager(self.logger)

    @info_function_start_finish('Setup Databases.')
    def run(self):
        # type: () -> None
        """
        Main sequence of events.
        """
        self.start_mongo_install_container()

    @verbose_function_start_finish('Start mongo installation container.')
    def start_mongo_install_container(self):
        # type: () -> None
        """
        Writes entered db encryption key to the file used in docker-compose
        """
        main_docker_compose = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_FILE)
        install_docker_compose = path.join(names.INSTALL_DIR, names.DOCKER_COMPOSE_INSTALL_FILE)
        docker_compose_manager = DockerComposeManager(self.logger, main_docker_compose, install_docker_compose)
        docker_compose_manager.docker_compose_run('--no-ansi', 'up', '-d', names.SW_MONGO)
        mongo_success_message = b"Successfully completed MongoDB initialization for Swimlane."
        error_message = b"ERROR: SWIMLANE INITIALIZATION ERROR OCCURRED"
        by_line = self.docker_manager.logs_by_line(names.SW_MONGO)
        try:
            start = time()
            for line in by_line:
                if mongo_success_message in line:
                    self.logger.info(str(mongo_success_message))
                    break
                if error_message in line:
                    raise ExpectedException('An error occurred initializing ${name}.'.format(name=names.SW_MONGO))
                if time() - start > 30:
                    raise ExpectedException('A timeout occurred initializing ${name}.'.format(name=names.SW_MONGO))
                else:
                    self.logger.debug(str(line))
        except ExpectedException:
            self.config.collect([{
                'type': 'confirm',
                'name': 'show_db_init_logs',
                'message': 'Error occurred, do you want to see the logs?'
            }])
            if self.config.args.show_db_init_logs:
                print(str(self.docker_manager.logs(names.SW_MONGO)))
            raise
        finally:
            by_line.close()
            docker_compose_manager.docker_compose_down()


def run(config):
    # type: (Configuration) -> None
    """
    The script run method, that can be called by other script or from file run method.
    :param config: Configuration information collected by parent script.
    """
    SetupDbs(config).run()



from pymongo import MongoClient
from typing import Union, Dict
from swimlane_platform.lib.db_configuration import DbConfiguration
from swimlane_platform.lib.args_config_questions import Configuration
from swimlane_platform.lib import BaseWithLog
from swimlane_platform.lib.expected_exceptions import ExpectedException
from swimlane_platform.lib.debug_decorators import debug_function_args, debug_function_return


class MongoManager(BaseWithLog):

    def __init__(self, config, uri=None):
        # type: (Configuration, Union[str, None]) -> None
        super(MongoManager, self).__init__(config)
        self.db_configuration = DbConfiguration(self.config)
        uri = uri if uri else self.get_admin_connection_uri()
        self.client = MongoClient(host=uri)

    @debug_function_return
    @debug_function_args
    def db_exists(self, db_name):
        # type: (str) -> bool
        """
        Verifies if database already exists on the server
        :param db_name: Database name
        :return: True if database exists, false otherwise.
        """
        try:
            existing_database_names = self.client.list_database_names()
        except Exception:
            raise ExpectedException('Cannot connect to database {name}'.format(name=db_name))
        return db_name in existing_database_names

    @debug_function_args
    def create_user(self, db_name, user_name, user_password, create_database=False, role='dbOwner'):
        # type: (str, str, str, bool, str) -> None
        """
        Creates the user in the selected database. Creates database if it doesn't exists and so requested.
        :param db_name: Name of the database.
        :param user_name: New user name.
        :param user_password: New user password.
        :param create_database: Should you create new database if it doesn't exists.
        :param role: The role user should have in the database.
        """
        try:
            existing_database_names = self.client.list_database_names()
        except Exception:
            raise ExpectedException('Cannot connect to database {name}'.format(name=db_name))

        if db_name not in existing_database_names and not create_database:
            raise ExpectedException('Database does not exist.')

        try:
            db = self.client.get_database(db_name)
        except Exception:
            raise ExpectedException('Cannot create database {name}'.format(name=db_name))

        try:
            db.command('createUser', user_name, pwd=user_password, roles=[{'role': role, 'db': db_name}])
        except Exception as e:
            raise ExpectedException('Cannot create user {name}. {error}'.format(name=user_name, error=e))

    @debug_function_return
    @debug_function_args
    def get_admin_connection_uri(self, ssl_req=True):
        # type: (bool) -> str
        """
        Gets default admin database connection URI.
        :param ssl_req: Is ssl required.
        :return: URI connection string.
        """
        uri_dict = dict()
        uri_dict['nodelist'] = [('localhost', 27017)]
        uri_dict['database'] = 'admin'
        uri_dict['password'] = self.db_configuration.get_admin_password()
        uri_dict['username'] = self.db_configuration.get_admin_user_name()
        if ssl_req:
            uri_dict['options'] = {'ssl': 'true', 'ssl_cert_reqs': 'CERT_NONE'}
        return self.db_configuration.get_mongo_uri(uri_dict)

    @debug_function_return
    @debug_function_args
    def user_exists(self, db, user_name):
        # type: (str, str) -> bool
        """
        Checks if user is defined in admin database for specific databases.
        :param db: Database name.
        :param user_name: User name. i.e. Swimlane.
        :return: True if user is defined.
        """
        users = self.client.get_database(db).command('usersInfo').get('users')
        return user_name in [user.get('user') for user in users]

    @debug_function_return
    @debug_function_args
    def get_swimlane_settings(self):
        # type: () -> Dict[str, Union[str, Dict]]
        """
        Returns current settings from Swimlane database.
        """
        database = self.client.get_database('Swimlane')
        collection = database.get_collection('Settings')
        return collection.find_one()

    @debug_function_return
    @debug_function_args
    def set_swimlane_settings(self, settings):
        # type: (object) -> None
        """
        Returns current settings from Swimlane database.
        """
        database = self.client.get_database('Swimlane')
        collection = database.get_collection('Settings')
        collection.find_one_and_replace({}, settings)

    @debug_function_args
    def set_database_compatibility(self, version):
        # type: (str) -> None
        """
        Sets compatibility option.
        :param version: Compatibility version.
        """
        self.client.get_database('admin').command('setFeatureCompatibilityVersion', version)



from __future__ import print_function
from swimlane_platform.lib import Configuration, names, DockerComposeFileManager, \
    BaseWithLog, info_function_start_finish, logging_questions
from os import path, makedirs
from uuid import uuid4


class AddFileEncryption(BaseWithLog):

    def __init__(self, config):
        # type: (Configuration) -> None
        super(AddFileEncryption, self).__init__(config)
        self.secrets = path.join(names.INSTALL_DIR, names.SECRETS_SUB_FOLDER)

    @info_function_start_finish('Add file encryption')
    def run(self):
        # type: () -> None
        """
        Runs modifications to environment to add file encryption key.
        """
        if not path.exists(self.secrets):
            makedirs(self.secrets)
        key = str(uuid4().hex) if self.config.args.autogenerated_file_encryption_key \
            else self.config.args.file_encryption_key
        key_file = path.join(self.secrets, 'file_encryption.key')
        if path.exists(key_file):
            self.logger.warn('File encryption key file already exists. We will not overwrite it.')
            return
        with open(key_file, 'w') as key_file_stream:
            key_file_stream.write(key)
        docker_compose = DockerComposeFileManager(self.logger, path.join(names.INSTALL_DIR, 'docker-compose.yml'))
        docker_compose.set(key_file, 'secrets', 'file_encryption_key', 'file')
        docker_compose.save()


def run(config):
    # type: (Configuration) -> None
    """
    The script run method, that can be called by other script or from file run method.
    :param config: Additional configuration information collected by parent script.
    """
    questions = [
        {
            'type': 'confirm',
            'message': 'Would you like to auto-generate a file encryption key in GUID format? If not, you must enter '
                       'your own file encryption key.',
            'name': 'autogenerated_file_encryption_key',
            'default': True
        },
        {
            'type': 'input',
            'name': 'file_encryption_key',
            'message': 'Please enter your file encryption key.',
            'when': lambda a: not a['autogenerated_file_encryption_key']
        }
    ]
    questions.extend(logging_questions)
    config.collect(questions)
    AddFileEncryption(config).run()



import logging
import sys


class LessThanFilter(logging.Filter):

    def __init__(self, exclusive_maximum, name=""):
        super(LessThanFilter, self).__init__(name)
        self.max_level = exclusive_maximum

    def filter(self, record):
        return 1 if record.levelno < self.max_level else 0


class SplitStreamLogger(logging.Logger):

    VERBOSE_LVL = 11

    def __init__(self, name, verbose, log_file=None):
        # type: (str, int, str) -> SplitStreamLogger
        super(SplitStreamLogger, self).__init__(name)
        logging.addLevelName(self.VERBOSE_LVL, 'VERBOSE')
        reporting_level = logging.INFO if verbose == 0 else self.VERBOSE_LVL if verbose == 1 else logging.DEBUG
        formatter = logging.Formatter('[%(asctime)s] %(levelname)s: %(message)s', '%Y-%m-%dT%H:%M:%S%z')
        log_out = logging.StreamHandler(sys.stdout)
        log_out.setLevel(reporting_level)
        log_out.addFilter(LessThanFilter(logging.WARNING))
        log_out.setFormatter(formatter)
        self.addHandler(log_out)
        log_error = logging.StreamHandler(sys.stderr)
        log_error.setLevel(logging.WARNING)
        log_error.setFormatter(formatter)
        self.addHandler(log_error)
        if log_file:
            self.addHandler(logging.FileHandler(log_file))

    def verbose(self, message, *args, **kws):
        if self.isEnabledFor(self.VERBOSE_LVL):
            message = str(message).replace("\n", "")
            self._log(self.VERBOSE_LVL, message, args, **kws)



import re
from typing import Union
from semantic_version import Version
from semver import VersionInfo


def pypi_version_to_semver(version):
    # type: (str) -> Union[VersionInfo, None]
    """
    Converts pypi version to VersionInfo if possible, if not returns None.
    :param version: PyPi version
    :return: VersionInfo object
    """
    suffix_expression = re.compile(r'(\.dev|[abc]|rc|\.post)\d+$')
    cleaned_version = suffix_expression.sub('', version)
    return semver_parse(cleaned_version)


def semver_parse(version_string):
    # type: (str) -> Union[VersionInfo, None]
    try:
        version = Version(version_string, partial=True)
        if not version.minor:
            version.minor = 0
        if not version.patch:
            version.patch = 0
        return VersionInfo.parse(str(version))
    except ValueError:
        return None



from OpenSSL import crypto
from OpenSSL.crypto import PKey, X509
from typing import Tuple


def create_dns_str(dns=""):
    # type: (str) -> [str]
    """
    Creates a Subject Alternative Name formatted string from a comma-separated string
    :return: comma-separated SAN formatted string
    """
    separator = ","
    formatted_dns = []
    for d in dns.split(separator):
        formatted_dns.append("DNS:" + d)
    return (separator + " ").join(formatted_dns)


def create_cert(country="US", state="Colorado", location="Lousville", company="Swimlane Inc", application="swimlane.io", dns="*.swimlane.io"):
    # type: (str, str, str, str, str, str) -> Tuple[X509, PKey]
    """
    Creates default self generated cert and key.
    :return: Tuple with cert and key
    """
    k = crypto.PKey()
    k.generate_key(crypto.TYPE_RSA, 2048)
    cert = crypto.X509()
    # this is actually X509v3, zero-based
    cert.set_version(2)
    cert.get_subject().C = country
    cert.get_subject().ST = state
    cert.get_subject().L = location
    cert.get_subject().OU = company
    cert.get_subject().CN = application
    cert.set_serial_number(1000)

    # validity period must be 825 days or fewer
    # https://support.apple.com/en-us/HT210176
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(824 * 24 * 60 * 60)

    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(k)

    # must contain subject alternative names with user provided dns entries
    # must contain EKU with id-kp-serverAuth OID
    # https://support.apple.com/en-us/HT210176
    formatted_dns = create_dns_str(dns)
    cert.add_extensions([
        crypto.X509Extension(b"subjectAltName", critical=True, value=formatted_dns.encode()),
        crypto.X509Extension(b"extendedKeyUsage", True, b"serverAuth")
    ])

    # must be SHA2+
    # https://support.apple.com/en-us/HT210176
    cert.sign(k, 'sha256')
    return cert, k
